# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


SCHEMA = {
    "PLUGIN_ACTIVATE": {"Implements": ["NetworkDriver", "IpamDriver"]},
    "SUCCESS": {}
}

# Routes are given a RouteType of 0 and a value for NextHop;
ROUTE_TYPE = {
    "NEXTHOP": 0
}

PROTOCOLS = {
    1: 'icmp',
    6: 'tcp',
    17: 'udp'
}

NET_NAME_PREFIX = 'kuryr-net-'
SUBNET_NAME_PREFIX = 'kuryr-subnet-'
NEUTRON_ID_LH_OPTION = 'kuryr.net.uuid.lh'
NEUTRON_ID_UH_OPTION = 'kuryr.net.uuid.uh'

DOCKER_EXPOSED_PORTS_OPTION = 'com.docker.network.endpoint.exposedports'
DOCKER_MAC_ADDRESS_OPTION = 'com.docker.network.endpoint.macaddress'
KURYR_EXISTING_NEUTRON_NET = 'kuryr.net.existing'
KURYR_EXISTING_NEUTRON_SUBNETPOOL = 'kuryr.subnetpool.existing'
KURYR_EXISTING_NEUTRON_PORT = 'kuryr.port.existing'
NETWORK_GATEWAY_OPTIONS = 'com.docker.network.gateway'
NETWORK_GENERIC_OPTIONS = 'com.docker.network.generic'
NEUTRON_NAME_OPTION = 'neutron.net.name'
NEUTRON_SHARED_OPTION = 'neutron.net.shared'
NEUTRON_SUBNET_NAME_OPTION = 'neutron.subnet.name'
NEUTRON_SUBNET_UUID_OPTION = 'neutron.subnet.uuid'
NEUTRON_V6_SUBNET_NAME_OPTION = 'neutron.subnet.v6.name'
NEUTRON_V6_SUBNET_UUID_OPTION = 'neutron.subnet.v6.uuid'
NEUTRON_POOL_NAME_OPTION = 'neutron.pool.name'
NEUTRON_POOL_UUID_OPTION = 'neutron.pool.uuid'
NEUTRON_V6_POOL_NAME_OPTION = 'neutron.pool.v6.name'
NEUTRON_V6_POOL_UUID_OPTION = 'neutron.pool.v6.uuid'
NEUTRON_UUID_OPTION = 'neutron.net.uuid'
REQUEST_ADDRESS_TYPE = 'RequestAddressType'
KURYR_UNBOUND_PORT = 'kuryr-unbound-port'
NEUTRON_UNBOUND_PORT = 'neutron-unbound-port'
BINDING_PROFILE = 'binding:profile'

# Define supported virtual NIC types.
VNIC_TYPE_NORMAL = 'normal'
VNIC_TYPE_DIRECT = 'direct'
VNIC_TYPE_MACVTAP = 'macvtap'
VNIC_TYPE_DIRECT_PHYSICAL = 'direct-physical'

# Define list of virtual NIC types.
VNIC_TYPES_SRIOV = (VNIC_TYPE_DIRECT, VNIC_TYPE_MACVTAP,
                    VNIC_TYPE_DIRECT_PHYSICAL)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import jsonschema
from jsonschema.exceptions import ValidationError

from kuryr_libnetwork import schemata
from kuryr_libnetwork.schemata import commons
from kuryr_libnetwork.tests.unit import base


class TestKuryrSchema(base.TestKuryrBase):
    """Unit tests for Kuryr schema."""

    def test_network_id_64_len(self):
        network_id = '51c75a2515d47edecc3f720bb541e287224416fb66715eb' \
                     '7802011d6ffd499f1'
        target_schema = commons.COMMONS[u'definitions'][u'id']
        self._validate_schema(network_id, target_schema)

    def test_network_id_25_len(self):
        network_id = 'xqqzd9p112o4kvok38n3caxjm'
        target_schema = commons.COMMONS[u'definitions'][u'id']
        self._validate_schema(network_id, target_schema)

    def test_network_id_invalid_charactor(self):
        network_id = '@#qzd9p112o4kvok38n3cax&%'
        target_schema = commons.COMMONS[u'definitions'][u'id']
        self.assertRaises(ValidationError, jsonschema.validate, network_id,
                          target_schema)

    def test_network_id_invalid_length(self):
        network_id = 'xqqzd9p112o4kvok38n3caxjmabcd'
        target_schema = commons.COMMONS[u'definitions'][u'id']
        self.assertRaises(ValidationError, jsonschema.validate, network_id,
                          target_schema)

    def test_network_create_schema(self):
        docker_network_id = '51c75a2515d47edecc3f720bb541e287224416fb66715eb' \
                            '7802011d6ffd499f1'
        network_request = {
            'NetworkID': docker_network_id,
            'IPv4Data': [{
                'AddressSpace': 'foo',
                'Pool': '192.168.42.0/24',
                'Gateway': '192.168.42.1/24',
                'AuxAddresses': {}
            }],
            'IPv6Data': [{
                'AddressSpace': 'bar',
                'Pool': 'fe80::/64',
                'Gateway': 'fe80::f816:3eff:fe20:57c3/64',
                'AuxAddresses': {}
            }],
            'Options': {}
        }
        self._validate_schema(network_request, schemata.NETWORK_CREATE_SCHEMA)

    def test_network_create_schema_missing_required(self):
        docker_network_id = '51c75a2515d47edecc3f720bb541e287224416fb66715eb' \
                            '7802011d6ffd499f1'
        net_request = {
            'NetworkID': docker_network_id,
            'IPv4Data': [{
                'AddressSpace': 'foo',
                'Pool': '192.168.42.0/24',
                'Gateway': '192.168.42.1/24',
                'AuxAddresses': {}
            }],
            'Options': {}
        }
        self.assertRaises(ValidationError, jsonschema.validate, net_request,
                          schemata.NETWORK_CREATE_SCHEMA)

    @classmethod
    def _validate_schema(self, target, schema):
        try:
            jsonschema.validate(target, schema)
        except ValidationError:
            self.fail("Unexpected validation error raised!")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from kuryr_libnetwork.schemata import commons

RELEASE_ADDRESS_SCHEMA = {
    u'links': [{
        u'method': u'POST',
        u'href': u'/IpamDriver.ReleaseAddress',
        u'description': u'Release an ip address',
        u'rel': u'self',
        u'title': u'Release'
    }],
    u'title': u'Release an IP',
    u'required': [u'PoolID', u'Address'],
    u'definitions': {u'commons': {}},
    u'$schema': u'http://json-schema.org/draft-04/hyper-schema',
    u'type': u'object',
    u'properties': {
        u'PoolID': {
            u'description': u'neutron uuid of allocated subnetpool',
            u'$ref': u'#/definitions/commons/definitions/uuid'
        },
        u'Address': {
            u'description': u'Address in IP(v4 or v6) form',
            u'$ref': u'#/definitions/commons/definitions/ipv4_or_ipv6'
        }
    }

}

RELEASE_ADDRESS_SCHEMA[u'definitions'][u'commons'] = commons.COMMONS



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ipaddress
from unittest import mock


import ddt
from oslo_serialization import jsonutils
from oslo_utils import uuidutils
from werkzeug import exceptions as w_exceptions

from kuryr.lib import constants as lib_const
from kuryr.lib import utils as lib_utils
from kuryr_libnetwork import config
from kuryr_libnetwork import constants as const
from kuryr_libnetwork.tests.unit import base
from kuryr_libnetwork import utils


FAKE_IP4_CIDR = '10.0.0.0/16'
FAKE_IP6_CIDR = 'fe80::/64'


@ddt.ddt
class TestKuryrIpam(base.TestKuryrBase):
    """Basic unit tests for libnetwork remote IPAM driver URI endpoints.

    This test class covers the following HTTP methods and URIs as described in
    the remote IPAM driver specification as below:

      https://github.com/docker/libnetwork/blob/9bf339f27e9f5c7c922036706c9bcc410899f249/docs/ipam.md  # noqa

    - POST /IpamDriver.GetDefaultAddressSpaces
    - POST /IpamDriver.RequestPool
    - POST /IpamDriver.ReleasePool
    - POST /IpamDriver.RequestAddress
    - POST /IpamDriver.ReleaseAddress
    """
    @ddt.data(
        ('/IpamDriver.GetDefaultAddressSpaces',
         {"LocalDefaultAddressSpace":
          config.CONF.local_default_address_space,
          "GlobalDefaultAddressSpace":
          config.CONF.global_default_address_space}),
        ('/IpamDriver.GetCapabilities',
         {"RequiresMACAddress": True}))
    @ddt.unpack
    def test_remote_ipam_driver_endpoint(self, endpoint, expected):
        response = self.app.post(endpoint)
        self.assertEqual(200, response.status_code)
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(expected, decoded_json)

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_subnetpool')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @ddt.data((FAKE_IP4_CIDR), (FAKE_IP6_CIDR))
    def test_ipam_driver_request_pool_with_existing_subnet_id(self,
            pool_cidr, mock_list_subnetpools,
            mock_create_subnetpool, mock_add_tag):
        neutron_subnet_v4_id = uuidutils.generate_uuid()

        pool_name = lib_utils.get_neutron_subnetpool_name(pool_cidr)
        prefixlen = ipaddress.ip_network(str(pool_cidr)).prefixlen
        new_subnetpool = {
            'name': pool_name,
            'default_prefixlen': prefixlen,
            'prefixes': [pool_cidr],
            'shared': False}

        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_name = pool_name
        if pool_cidr == FAKE_IP4_CIDR:
            kuryr_subnetpools = self._get_fake_v4_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr],
                name=fake_name)
        else:
            kuryr_subnetpools = self._get_fake_v6_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr],
                name=fake_name)
        mock_list_subnetpools.return_value = {'subnetpools': []}
        fake_subnetpool_response = {
            'subnetpool': kuryr_subnetpools['subnetpools'][0]
        }

        mock_create_subnetpool.return_value = fake_subnetpool_response

        fake_request = {
            'AddressSpace': '',
            'Pool': pool_cidr,
            'SubPool': '',  # In the case --ip-range is not given
            'Options': {
                'neutron.subnet.uuid': neutron_subnet_v4_id
            },
            'V6': False
        }
        response = self.app.post('/IpamDriver.RequestPool',
                                 content_type='application/json',
                                 data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnetpools.assert_called_with(
            name=pool_name, tags=[str(neutron_subnet_v4_id)])
        mock_create_subnetpool.assert_called_with(
            {'subnetpool': new_subnetpool})
        mock_add_tag.assert_called_once_with(
            'subnetpools', fake_kuryr_subnetpool_id, neutron_subnet_v4_id)
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(fake_kuryr_subnetpool_id, decoded_json['PoolID'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_subnetpool')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @ddt.data((FAKE_IP4_CIDR), (FAKE_IP6_CIDR))
    def test_ipam_driver_request_pool_with_existing_subnet_id_and_shared(self,
            pool_cidr, mock_list_subnetpools,
            mock_create_subnetpool, mock_add_tag):
        neutron_subnet_v4_id = uuidutils.generate_uuid()

        pool_name = lib_utils.get_neutron_subnetpool_name(pool_cidr)
        prefixlen = ipaddress.ip_network(str(pool_cidr)).prefixlen
        new_subnetpool = {
            'name': pool_name,
            'default_prefixlen': prefixlen,
            'prefixes': [pool_cidr],
            'shared': True}

        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_name = pool_name
        if pool_cidr == FAKE_IP4_CIDR:
            kuryr_subnetpools = self._get_fake_v4_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr],
                name=fake_name)
        else:
            kuryr_subnetpools = self._get_fake_v6_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr],
                name=fake_name)
        mock_list_subnetpools.return_value = {'subnetpools': []}
        fake_subnetpool_response = {
            'subnetpool': kuryr_subnetpools['subnetpools'][0]
        }

        mock_create_subnetpool.return_value = fake_subnetpool_response

        fake_request = {
            'AddressSpace': '',
            'Pool': pool_cidr,
            'SubPool': '',  # In the case --ip-range is not given
            'Options': {
                'neutron.subnet.uuid': neutron_subnet_v4_id,
                'neutron.net.shared': True
            },
            'V6': False
        }
        response = self.app.post('/IpamDriver.RequestPool',
                                 content_type='application/json',
                                 data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnetpools.assert_called_with(
            name=pool_name, tags=[str(neutron_subnet_v4_id)])
        mock_create_subnetpool.assert_called_with(
            {'subnetpool': new_subnetpool})
        mock_add_tag.assert_called_once_with(
            'subnetpools', fake_kuryr_subnetpool_id, neutron_subnet_v4_id)
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(fake_kuryr_subnetpool_id, decoded_json['PoolID'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_subnetpool')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @ddt.data((FAKE_IP4_CIDR), (FAKE_IP6_CIDR))
    def test_ipam_driver_request_pool_with_existing_subnet_name(self,
            pool_cidr, mock_list_subnets, mock_list_subnetpools,
            mock_create_subnetpool, mock_add_tag):
        # faking list_subnets
        docker_endpoint_id = lib_utils.get_hash()
        neutron_network_id = uuidutils.generate_uuid()
        neutron_subnet_v4_id = uuidutils.generate_uuid()
        neutron_subnet_v4_name = utils.make_subnet_name(FAKE_IP4_CIDR)

        # Faking existing Neutron subnets
        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id,
            subnet_v4_id=neutron_subnet_v4_id,
            cidr=FAKE_IP4_CIDR, name=neutron_subnet_v4_name)

        fake_subnets = {
            'subnets': [
                fake_v4_subnet['subnet'],
            ]
        }

        mock_list_subnets.return_value = fake_subnets
        pool_name = lib_utils.get_neutron_subnetpool_name(pool_cidr)
        prefixlen = ipaddress.ip_network(str(pool_cidr)).prefixlen
        new_subnetpool = {
            'name': pool_name,
            'default_prefixlen': prefixlen,
            'prefixes': [pool_cidr],
            'shared': False}

        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_name = pool_name
        if pool_cidr == FAKE_IP4_CIDR:
            kuryr_subnetpools = self._get_fake_v4_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr],
                name=fake_name)
        else:
            kuryr_subnetpools = self._get_fake_v6_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr],
                name=fake_name)
        mock_list_subnetpools.return_value = {'subnetpools': []}
        fake_subnetpool_response = {
            'subnetpool': kuryr_subnetpools['subnetpools'][0]
        }

        mock_create_subnetpool.return_value = fake_subnetpool_response

        fake_request = {
            'AddressSpace': '',
            'Pool': pool_cidr,
            'SubPool': '',  # In the case --ip-range is not given
            'Options': {
                'neutron.subnet.name': neutron_subnet_v4_name
            },
            'V6': False
        }
        response = self.app.post('/IpamDriver.RequestPool',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(name=neutron_subnet_v4_name)
        mock_list_subnetpools.assert_called_with(
            name=pool_name, tags=[str(neutron_subnet_v4_id)])
        mock_create_subnetpool.assert_called_with(
            {'subnetpool': new_subnetpool})
        mock_add_tag.assert_called_once_with(
            'subnetpools', fake_kuryr_subnetpool_id, neutron_subnet_v4_id)
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(fake_kuryr_subnetpool_id, decoded_json['PoolID'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_subnetpool')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @ddt.data((FAKE_IP4_CIDR), (FAKE_IP6_CIDR))
    def test_ipam_driver_request_pool_with_user_pool(self, pool_cidr,
            mock_list_subnets, mock_list_subnetpools, mock_create_subnetpool):
        fake_subnet = {"subnets": []}
        mock_list_subnets.return_value = fake_subnet
        pool_name = lib_utils.get_neutron_subnetpool_name(pool_cidr)
        prefixlen = ipaddress.ip_network(str(pool_cidr)).prefixlen
        new_subnetpool = {
            'name': pool_name,
            'default_prefixlen': prefixlen,
            'prefixes': [pool_cidr],
            'shared': False}

        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_name = pool_name
        if pool_cidr == FAKE_IP4_CIDR:
            kuryr_subnetpools = self._get_fake_v4_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr],
                name=fake_name)
        else:
            kuryr_subnetpools = self._get_fake_v6_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr],
                name=fake_name)
        mock_list_subnetpools.return_value = {'subnetpools': []}
        fake_subnetpool_response = {
            'subnetpool': kuryr_subnetpools['subnetpools'][0]
        }

        mock_create_subnetpool.return_value = fake_subnetpool_response

        fake_request = {
            'AddressSpace': '',
            'Pool': pool_cidr,
            'SubPool': '',  # In the case --ip-range is not given
            'Options': {},
            'V6': False
        }
        response = self.app.post('/IpamDriver.RequestPool',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(cidr=pool_cidr)
        mock_list_subnetpools.assert_called_with(name=fake_name)
        mock_create_subnetpool.assert_called_with(
            {'subnetpool': new_subnetpool})
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(fake_kuryr_subnetpool_id, decoded_json['PoolID'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((True, FAKE_IP4_CIDR), (True, FAKE_IP6_CIDR),
              (False, FAKE_IP4_CIDR), (False, FAKE_IP6_CIDR))
    @ddt.unpack
    def test_ipam_driver_request_pool_with_pool_name_option(self,
            use_tag_ext, pool_cidr, mock_app, mock_list_subnets,
            mock_list_subnetpools, mock_add_tag):
        mock_app.tag_ext = use_tag_ext
        fake_subnet = {"subnets": []}
        mock_list_subnets.return_value = fake_subnet

        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_name = 'fake_pool_name'
        if pool_cidr == FAKE_IP4_CIDR:
            kuryr_subnetpools = self._get_fake_v4_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr],
                name=fake_name)
            options = {
                const.NEUTRON_POOL_NAME_OPTION: fake_name}
        else:
            kuryr_subnetpools = self._get_fake_v6_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr],
                name=fake_name)
            options = {
                const.NEUTRON_V6_POOL_NAME_OPTION: fake_name}
        mock_list_subnetpools.return_value = kuryr_subnetpools

        fake_request = {
            'AddressSpace': '',
            'Pool': pool_cidr,
            'SubPool': pool_cidr,
            'Options': options,
            'V6': False if pool_cidr == FAKE_IP4_CIDR else True
        }
        response = self.app.post('/IpamDriver.RequestPool',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(cidr=pool_cidr)
        if mock_app.tag_ext:
            mock_add_tag.assert_called_once_with(
                'subnetpools', fake_kuryr_subnetpool_id,
                const.KURYR_EXISTING_NEUTRON_SUBNETPOOL)
        else:
            mock_add_tag.assert_not_called()
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(fake_kuryr_subnetpool_id, decoded_json['PoolID'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((True, FAKE_IP4_CIDR), (True, FAKE_IP6_CIDR),
              (False, FAKE_IP4_CIDR), (False, FAKE_IP6_CIDR))
    @ddt.unpack
    def test_ipam_driver_request_pool_with_pool_id_option(self,
            use_tag_ext, pool_cidr, mock_app, mock_list_subnets,
            mock_list_subnetpools, mock_add_tag):
        mock_app.tag_ext = use_tag_ext
        fake_subnet = {"subnets": []}
        mock_list_subnets.return_value = fake_subnet

        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        if pool_cidr == FAKE_IP4_CIDR:
            kuryr_subnetpools = self._get_fake_v4_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr])
            options = {
                const.NEUTRON_POOL_UUID_OPTION: fake_kuryr_subnetpool_id}
        else:
            kuryr_subnetpools = self._get_fake_v6_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[pool_cidr])
            options = {
                const.NEUTRON_V6_POOL_UUID_OPTION: fake_kuryr_subnetpool_id}
        mock_list_subnetpools.return_value = kuryr_subnetpools

        fake_request = {
            'AddressSpace': '',
            'Pool': pool_cidr,
            'SubPool': pool_cidr,
            'Options': options,
            'V6': False if pool_cidr == FAKE_IP4_CIDR else True
        }
        response = self.app.post('/IpamDriver.RequestPool',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(cidr=pool_cidr)
        if mock_app.tag_ext:
            mock_add_tag.assert_called_once_with(
                'subnetpools', fake_kuryr_subnetpool_id,
                const.KURYR_EXISTING_NEUTRON_SUBNETPOOL)
        else:
            mock_add_tag.assert_not_called()
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(fake_kuryr_subnetpool_id, decoded_json['PoolID'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_subnetpool')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((True, FAKE_IP4_CIDR), (True, FAKE_IP6_CIDR),
              (False, FAKE_IP4_CIDR), (False, FAKE_IP6_CIDR))
    @ddt.unpack
    def test_ipam_driver_request_pool_with_unmatched_cidr(self,
            use_tag_ext, pool_cidr, mock_app, mock_list_subnets,
            mock_list_subnetpools, mock_create_subnetpool, mock_add_tag):
        mock_app.tag_ext = use_tag_ext
        fake_subnet = {"subnets": []}
        mock_list_subnets.return_value = fake_subnet
        subnet_ip4_cidr = '10.0.0.0/24'
        subnet_ip6_cidr = 'fe80::/68'
        if pool_cidr == FAKE_IP4_CIDR:
            subnet_cidr = subnet_ip4_cidr
        else:
            subnet_cidr = subnet_ip6_cidr
        pool_name = lib_utils.get_neutron_subnetpool_name(subnet_cidr)
        prefixlen = ipaddress.ip_network(str(subnet_cidr)).prefixlen
        new_subnetpool = {
            'name': pool_name,
            'default_prefixlen': prefixlen,
            'prefixes': [subnet_cidr],
            'shared': False}

        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_existing_subnetpool_id = uuidutils.generate_uuid()
        if pool_cidr == FAKE_IP4_CIDR:
            kuryr_subnetpools = self._get_fake_v4_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[subnet_ip4_cidr])
            existing_subnetpools = self._get_fake_v4_subnetpools(
                fake_existing_subnetpool_id, prefixes=[pool_cidr])
            options = {
                const.NEUTRON_POOL_UUID_OPTION: fake_existing_subnetpool_id}
        else:
            kuryr_subnetpools = self._get_fake_v6_subnetpools(
                fake_kuryr_subnetpool_id, prefixes=[subnet_ip6_cidr])
            existing_subnetpools = self._get_fake_v6_subnetpools(
                fake_existing_subnetpool_id, prefixes=[pool_cidr])
            options = {
                const.NEUTRON_V6_POOL_UUID_OPTION: fake_existing_subnetpool_id}
        mock_list_subnetpools.side_effect = [
            existing_subnetpools,
            {'subnetpools': []}
        ]

        fake_subnetpool_response = {
            'subnetpool': kuryr_subnetpools['subnetpools'][0]
        }
        mock_create_subnetpool.return_value = fake_subnetpool_response

        if pool_cidr == FAKE_IP4_CIDR:
            subnet_cidr = subnet_ip4_cidr
            fake_request = {
                'AddressSpace': '',
                'Pool': subnet_cidr,
                'SubPool': subnet_cidr,
                'Options': options,
                'V6': False,
            }
        else:
            subnet_cidr = subnet_ip6_cidr
            fake_request = {
                'AddressSpace': '',
                'Pool': subnet_cidr,
                'SubPool': subnet_cidr,
                'Options': options,
                'V6': True,
            }
        response = self.app.post('/IpamDriver.RequestPool',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(cidr=subnet_cidr)
        mock_create_subnetpool.assert_called_with(
            {'subnetpool': new_subnetpool})
        if mock_app.tag_ext:
            mock_add_tag.assert_called_once_with(
                'subnetpools', fake_existing_subnetpool_id,
                const.KURYR_EXISTING_NEUTRON_SUBNETPOOL)
        else:
            mock_add_tag.assert_not_called()
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(fake_kuryr_subnetpool_id, decoded_json['PoolID'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    def test_ipam_driver_request_pool_with_default_v6pool(self,
            mock_list_subnetpools):
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_name = 'kuryr6'
        kuryr_subnetpools = self._get_fake_v6_subnetpools(
            fake_kuryr_subnetpool_id, prefixes=['fe80::/64'])
        mock_list_subnetpools.return_value = {
            'subnetpools': kuryr_subnetpools['subnetpools']}

        fake_request = {
            'AddressSpace': '',
            'Pool': '',
            'SubPool': '',  # In the case --ip-range is not given
            'Options': {},
            'V6': True
        }
        response = self.app.post('/IpamDriver.RequestPool',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnetpools.assert_called_with(name=fake_name)
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(fake_kuryr_subnetpool_id, decoded_json['PoolID'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.delete_subnetpool')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.remove_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((True), (False))
    def test_ipam_driver_release_pool(self,
                                      use_tag_ext,
                                      mock_app,
                                      mock_list_subnetpools,
                                      mock_list_subnets,
                                      mock_remove_tag,
                                      mock_delete_subnetpool):
        mock_app.tag_ext = use_tag_ext

        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_subnetpool_name = lib_utils.get_neutron_subnetpool_name(
            FAKE_IP4_CIDR)
        kuryr_subnetpools = self._get_fake_v4_subnetpools(
            fake_kuryr_subnetpool_id, prefixes=[FAKE_IP4_CIDR],
            name=fake_subnetpool_name)
        mock_list_subnetpools.return_value = kuryr_subnetpools

        docker_endpoint_id = lib_utils.get_hash()
        neutron_network_id = uuidutils.generate_uuid()
        subnet_v4_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.return_value = fake_subnet_response

        mock_delete_subnetpool.return_value = {}

        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id
        }
        response = self.app.post('/IpamDriver.ReleasePool',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        if mock_app.tag_ext:
            mock_list_subnetpools.assert_called_with(
                id=fake_kuryr_subnetpool_id)
            mock_list_subnets.assert_any_call(
                cidr=FAKE_IP4_CIDR)
            mock_remove_tag.assert_called_with('subnets',
                                               subnet_v4_id,
                                               fake_kuryr_subnetpool_id)
        mock_delete_subnetpool.assert_called_with(fake_kuryr_subnetpool_id)

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.remove_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((True), (False))
    def test_ipam_driver_release_pool_with_pool_name_option(
            self, use_tag_ext, mock_app, mock_list_subnetpools,
            mock_list_subnets, mock_remove_tag):
        mock_app.tag_ext = use_tag_ext

        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_subnetpool_name = 'fake_pool_name'
        fake_tags = []
        if mock_app.tag_ext:
            fake_tags.append(const.KURYR_EXISTING_NEUTRON_SUBNETPOOL)
        kuryr_subnetpools = self._get_fake_v4_subnetpools(
            fake_kuryr_subnetpool_id, prefixes=[FAKE_IP4_CIDR],
            name=fake_subnetpool_name, tags=fake_tags)
        mock_list_subnetpools.return_value = kuryr_subnetpools

        docker_endpoint_id = lib_utils.get_hash()
        neutron_network_id = uuidutils.generate_uuid()
        subnet_v4_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.return_value = fake_subnet_response

        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id
        }
        response = self.app.post('/IpamDriver.ReleasePool',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        if mock_app.tag_ext:
            mock_list_subnetpools.assert_called_with(
                id=fake_kuryr_subnetpool_id)
            mock_list_subnets.assert_called_with(
                cidr=FAKE_IP4_CIDR)
            mock_remove_tag.assert_any_call('subnets',
                                            subnet_v4_id,
                                            fake_kuryr_subnetpool_id)
            mock_remove_tag.assert_any_call(
                'subnetpools', fake_kuryr_subnetpool_id,
                const.KURYR_EXISTING_NEUTRON_SUBNETPOOL)

    @mock.patch('kuryr_libnetwork.controllers._neutron_port_add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((False), (True))
    def test_ipam_driver_request_address(self, use_tag_ext, mock_app,
            mock_list_subnets, mock_create_port, mock_port_add_tag):
        mock_app.tag_ext = use_tag_ext
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        # faking list_subnets
        docker_endpoint_id = lib_utils.get_hash()
        neutron_network_id = uuidutils.generate_uuid()
        subnet_v4_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.return_value = fake_subnet_response

        # faking create_port
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_mac_address = 'fa:16:3e:ca:59:88'
        fake_port = base.TestKuryrBase._get_fake_port(
            docker_endpoint_id, neutron_network_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            subnet_v4_id,
            neutron_subnet_v4_address="10.0.0.5",
            neutron_mac_address=fake_mac_address)
        port_request = {
            'name': const.KURYR_UNBOUND_PORT,
            'admin_state_up': True,
            'network_id': neutron_network_id,
        }
        fixed_ips = port_request['fixed_ips'] = []
        fixed_ip = {'subnet_id': subnet_v4_id}
        fixed_ips.append(fixed_ip)

        mock_create_port.return_value = fake_port
        # Testing container ip allocation
        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': '',  # Querying for container address
            'Options': {const.DOCKER_MAC_ADDRESS_OPTION: fake_mac_address}
        }
        mock_port_add_tag.return_value = None
        response = self.app.post('/IpamDriver.RequestAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(
            subnetpool_id=fake_kuryr_subnetpool_id)
        mock_create_port.assert_called_with({'port': port_request})
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual('10.0.0.5/16', decoded_json['Address'])
        if mock_app.tag_ext:
            mock_port_add_tag.assert_called()
        else:
            mock_port_add_tag.assert_not_called()

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    def test_ipam_driver_request_address_when_subnet_not_exist(self,
            mock_list_subnetpools, mock_list_subnets):
        requested_address = '10.0.0.5'
        fake_mac_address = 'fa:16:3e:ca:59:88'
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_name = lib_utils.get_neutron_subnetpool_name(FAKE_IP4_CIDR)
        kuryr_subnetpools = self._get_fake_v4_subnetpools(
            fake_kuryr_subnetpool_id, prefixes=[FAKE_IP4_CIDR],
            name=fake_name)
        mock_list_subnetpools.return_value = kuryr_subnetpools

        # faking list_subnets
        fake_subnet_response = {'subnets': []}
        mock_list_subnets.return_value = fake_subnet_response

        # Testing container ip allocation
        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': requested_address,
            'Options': {const.DOCKER_MAC_ADDRESS_OPTION: fake_mac_address}
        }
        response = self.app.post('/IpamDriver.RequestAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(500, response.status_code)
        mock_list_subnetpools.assert_called_with(
            id=fake_kuryr_subnetpool_id)
        mock_list_subnets.assert_called_with(cidr=FAKE_IP4_CIDR)

    @mock.patch('kuryr_libnetwork.controllers._neutron_port_add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.update_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((False), (True))
    def test_ipam_driver_request_specific_address(self,
            use_tag_ext, mock_app, mock_list_subnets, mock_list_ports,
            mock_update_port, mock_create_port, mock_port_add_tag):
        mock_app.tag_ext = use_tag_ext
        # faking list_subnets
        neutron_network_id = uuidutils.generate_uuid()
        docker_endpoint_id = lib_utils.get_hash()
        subnet_v4_id = uuidutils.generate_uuid()
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.return_value = fake_subnet_response

        # faking update_port or create_port
        requested_address = '10.0.0.5'
        fake_mac_address = 'fa:16:3e:ca:59:88'
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_port = base.TestKuryrBase._get_fake_port(
            docker_endpoint_id, neutron_network_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            subnet_v4_id,
            neutron_subnet_v4_address=requested_address,
            neutron_mac_address=fake_mac_address)

        fixed_ip_existing = [('subnet_id=%s' % subnet_v4_id)]
        fixed_ip_existing.append('ip_address=%s' % requested_address)
        fake_ports_response = {'ports': []}
        mock_list_ports.return_value = fake_ports_response

        port_request = {
            'name': const.KURYR_UNBOUND_PORT,
            'admin_state_up': True,
            'network_id': neutron_network_id,
        }
        fixed_ips = port_request['fixed_ips'] = []
        fixed_ip = {'subnet_id': subnet_v4_id,
                    'ip_address': requested_address}
        fixed_ips.append(fixed_ip)
        mock_create_port.return_value = fake_port

        # Testing container ip allocation
        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': requested_address,
            'Options': {const.DOCKER_MAC_ADDRESS_OPTION: fake_mac_address}
        }
        mock_port_add_tag.return_value = None
        response = self.app.post('/IpamDriver.RequestAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(
            subnetpool_id=fake_kuryr_subnetpool_id)
        mock_list_ports.assert_has_calls([
            mock.call(fixed_ips=fixed_ip_existing),
            mock.call(
                mac_address=fake_mac_address,
                fixed_ips='subnet_id=%s' % fake_v4_subnet['subnet']['id'])])
        mock_create_port.assert_called_with({'port': port_request})
        if mock_app.tag_ext:
            mock_port_add_tag.assert_called()
        else:
            mock_port_add_tag.assert_not_called()
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(requested_address + '/16', decoded_json['Address'])

    @mock.patch('kuryr_libnetwork.controllers._neutron_port_add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((False), (True))
    def test_ipam_driver_request_specific_address_existing_port(self,
            use_tag_ext, mock_app, mock_list_subnets, mock_list_ports,
            mock_create_port, mock_port_add_tag):
        mock_app.tag_ext = use_tag_ext
        # faking list_subnets
        neutron_network_id = uuidutils.generate_uuid()
        docker_endpoint_id = lib_utils.get_hash()
        subnet_v4_id = uuidutils.generate_uuid()
        subnet_v6_id = uuidutils.generate_uuid()
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_kuryr_subnetpool_v6_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_v6_subnet = self._get_fake_v6_subnet(
            neutron_network_id, docker_endpoint_id, subnet_v6_id,
            subnetpool_id=fake_kuryr_subnetpool_v6_id,
            cidr=FAKE_IP6_CIDR)
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        fake_subnet_response_v6 = {
            'subnets': [
                fake_v6_subnet['subnet']
            ]
        }
        mock_list_subnets.side_effect = [
            fake_subnet_response, fake_subnet_response_v6]

        # faking update_port or create_port
        requested_address = '10.0.0.5'
        requested_address_v6 = 'fe80::6'
        requested_mac_address = 'fa:16:3e:86:a0:fe'
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_port = base.TestKuryrBase._get_fake_port(
            docker_endpoint_id, neutron_network_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            subnet_v4_id, subnet_v6_id,
            neutron_subnet_v4_address=requested_address,
            neutron_subnet_v6_address=requested_address_v6,
            neutron_mac_address=requested_mac_address)

        fixed_ip_existing = [('subnet_id=%s' % subnet_v4_id)]
        fixed_ipv6_existing = [('subnet_id=%s' % subnet_v6_id)]
        fixed_ip_existing.append('ip_address=%s' % requested_address)
        fixed_ipv6_existing.append('ip_address=%s' % requested_address_v6)
        fake_existing_port = dict(fake_port['port'])
        fake_existing_port['binding:host_id'] = ''
        fake_existing_port['binding:vif_type'] = 'unbound'
        fake_ports_response = {'ports': [fake_existing_port]}
        fake_existing_port_2 = dict(fake_port['port'])
        fake_existing_port_2['name'] = const.NEUTRON_UNBOUND_PORT
        fake_existing_port_2['binding:host_id'] = lib_utils.get_hostname()
        fake_ports_response_2 = {'ports': [fake_existing_port_2]}
        mock_list_ports.side_effect = [
            fake_ports_response, fake_ports_response_2]

        # Testing container ip allocation
        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': requested_address,
            'Options': {const.DOCKER_MAC_ADDRESS_OPTION: requested_mac_address}
        }
        mock_port_add_tag.return_value = None
        response = self.app.post('/IpamDriver.RequestAddress',
                                 content_type='application/json',
                                 data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(requested_address + '/16', decoded_json['Address'])

        fake_request_2 = {
            'PoolID': fake_kuryr_subnetpool_v6_id,
            'Address': requested_address_v6,
            'Options': {const.DOCKER_MAC_ADDRESS_OPTION: requested_mac_address}
        }
        response = self.app.post('/IpamDriver.RequestAddress',
                                 content_type='application/json',
                                 data=jsonutils.dumps(fake_request_2))

        self.assertEqual(200, response.status_code)
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(requested_address_v6 + '/64', decoded_json['Address'])

        mock_list_subnets.assert_has_calls([
            mock.call(subnetpool_id=fake_kuryr_subnetpool_id),
            mock.call(subnetpool_id=fake_kuryr_subnetpool_v6_id)])
        mock_list_ports.assert_has_calls([
            mock.call(fixed_ips=fixed_ip_existing),
            mock.call(fixed_ips=fixed_ipv6_existing)])
        if mock_app.tag_ext:
            self.assertEqual(2, mock_port_add_tag.call_count)
        else:
            self.assertEqual(0, mock_port_add_tag.call_count)

    @mock.patch('kuryr_libnetwork.controllers._neutron_port_add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((False), (True))
    def test_ipam_driver_request_specific_mac_address_existing_port(self,
            use_tag_ext, mock_app, mock_list_subnets, mock_list_ports,
            mock_create_port, mock_port_add_tag):
        mock_app.tag_ext = use_tag_ext
        # faking list_subnets
        neutron_network_id = uuidutils.generate_uuid()
        docker_endpoint_id = lib_utils.get_hash()
        subnet_v4_id = uuidutils.generate_uuid()
        subnet_v6_id = uuidutils.generate_uuid()
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_kuryr_subnetpool_v6_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_v6_subnet = self._get_fake_v6_subnet(
            neutron_network_id, docker_endpoint_id, subnet_v6_id,
            subnetpool_id=fake_kuryr_subnetpool_v6_id,
            cidr=FAKE_IP6_CIDR)
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        fake_subnet_response_v6 = {
            'subnets': [
                fake_v6_subnet['subnet']
            ]
        }
        mock_list_subnets.side_effect = [
            fake_subnet_response, fake_subnet_response_v6]

        # faking update_port or create_port
        fake_address = '10.0.0.5'
        fake_address_v6 = 'fe80::6'
        requested_mac_address = 'fa:16:3e:86:a0:fe'
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_port = base.TestKuryrBase._get_fake_port(
            docker_endpoint_id, neutron_network_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            subnet_v4_id, subnet_v6_id,
            neutron_subnet_v4_address=fake_address,
            neutron_subnet_v6_address=fake_address_v6,
            neutron_mac_address=requested_mac_address)

        fixed_ip_existing = [('subnet_id=%s' % subnet_v4_id)]
        fixed_ipv6_existing = [('subnet_id=%s' % subnet_v6_id)]
        fixed_ip_existing.append('ip_address=%s' % fake_address)
        fixed_ipv6_existing.append('ip_address=%s' % fake_address_v6)
        fake_existing_port = dict(fake_port['port'])
        fake_existing_port['binding:host_id'] = ''
        fake_existing_port['binding:vif_type'] = 'unbound'
        fake_ports_response = {'ports': [fake_existing_port]}
        fake_existing_port_2 = dict(fake_port['port'])
        fake_existing_port_2['name'] = const.NEUTRON_UNBOUND_PORT
        fake_existing_port_2['binding:host_id'] = lib_utils.get_hostname()
        fake_ports_response_2 = {'ports': [fake_existing_port_2]}
        mock_list_ports.side_effect = [
            fake_ports_response, fake_ports_response_2]

        # Testing container ip allocation
        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': '',
            'Options': {const.DOCKER_MAC_ADDRESS_OPTION: requested_mac_address}
        }
        mock_port_add_tag.return_value = None
        response = self.app.post('/IpamDriver.RequestAddress',
                                 content_type='application/json',
                                 data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(fake_address + '/16', decoded_json['Address'])

        fake_request_2 = {
            'PoolID': fake_kuryr_subnetpool_v6_id,
            'Address': '',
            'Options': {const.DOCKER_MAC_ADDRESS_OPTION: requested_mac_address}
        }
        response = self.app.post('/IpamDriver.RequestAddress',
                                 content_type='application/json',
                                 data=jsonutils.dumps(fake_request_2))

        self.assertEqual(200, response.status_code)
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(fake_address_v6 + '/64', decoded_json['Address'])

        mock_list_subnets.assert_has_calls([
            mock.call(subnetpool_id=fake_kuryr_subnetpool_id),
            mock.call(subnetpool_id=fake_kuryr_subnetpool_v6_id)])
        mock_list_ports.assert_has_calls([
            mock.call(mac_address=requested_mac_address,
                      fixed_ips='subnet_id=%s' % subnet_v4_id),
            mock.call(mac_address=requested_mac_address,
                      fixed_ips='subnet_id=%s' % subnet_v6_id)])
        if mock_app.tag_ext:
            self.assertEqual(2, mock_port_add_tag.call_count)
        else:
            self.assertEqual(0, mock_port_add_tag.call_count)

    @mock.patch('kuryr_libnetwork.controllers._neutron_port_add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((False), (True))
    def test_ipam_driver_request_address_overlapping_cidr_in_neutron(self,
            use_tag_ext, mock_app, mock_list_subnetpools, mock_list_subnets,
            mock_create_port, mock_port_add_tag):
        mock_app.tag_ext = use_tag_ext
        # faking list_subnetpools
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_kuryr_subnetpool_id2 = uuidutils.generate_uuid()

        fake_name = lib_utils.get_neutron_subnetpool_name(FAKE_IP4_CIDR)
        kuryr_subnetpools = self._get_fake_v4_subnetpools(
            fake_kuryr_subnetpool_id, prefixes=[FAKE_IP4_CIDR],
            name=fake_name)
        mock_list_subnetpools.return_value = kuryr_subnetpools

        # faking list_subnets
        docker_endpoint_id = lib_utils.get_hash()
        neutron_network_id = uuidutils.generate_uuid()
        neutron_network_id2 = uuidutils.generate_uuid()
        neutron_subnet_v4_id = uuidutils.generate_uuid()
        neutron_subnet_v4_id2 = uuidutils.generate_uuid()

        # Fake existing Neutron subnets
        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id, neutron_subnet_v4_id,
            cidr=FAKE_IP4_CIDR, name=utils.make_subnet_name(FAKE_IP4_CIDR),
            tags=[fake_kuryr_subnetpool_id])

        fake_v4_subnet2 = self._get_fake_v4_subnet(
            neutron_network_id2, docker_endpoint_id, neutron_subnet_v4_id2,
            cidr=FAKE_IP4_CIDR, name=utils.make_subnet_name(FAKE_IP4_CIDR),
            tags=[fake_kuryr_subnetpool_id2])

        fake_subnet_response = {
            'subnets': []
        }
        fake_subnet_response2 = {
            'subnets': [
                fake_v4_subnet2['subnet'],
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.side_effect = [
            fake_subnet_response, fake_subnet_response2]
        # faking create_port
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_mac_address = 'fa:16:3e:86:a0:fe'
        fake_port = self._get_fake_port(
            docker_endpoint_id, neutron_network_id,
            fake_neutron_port_id,
            neutron_subnet_v4_id=neutron_subnet_v4_id,
            neutron_subnet_v4_address="10.0.0.5",
            neutron_mac_address=fake_mac_address)
        mock_create_port.return_value = fake_port
        port_request = {
            'name': const.KURYR_UNBOUND_PORT,
            'admin_state_up': True,
            'network_id': neutron_network_id,
        }
        port_request['fixed_ips'] = []
        fixed_ip = {'subnet_id': neutron_subnet_v4_id}
        port_request['fixed_ips'].append(fixed_ip)

        # Testing container ip allocation
        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': '',  # Querying for container address
            'Options': {const.DOCKER_MAC_ADDRESS_OPTION: fake_mac_address}
        }
        mock_port_add_tag.return_value = None
        response = self.app.post('/IpamDriver.RequestAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnetpools.assert_called_with(
            id=fake_kuryr_subnetpool_id)
        mock_list_subnets.assert_called_with(
            cidr=FAKE_IP4_CIDR)
        mock_create_port.assert_called_with(
            {'port': port_request})
        if mock_app.tag_ext:
            mock_port_add_tag.assert_called()
        else:
            mock_port_add_tag.assert_not_called()
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual('10.0.0.5/16', decoded_json['Address'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    def test_ipam_driver_request_address_overlapping_cidr_no_subnet_tags(self,
            mock_list_subnetpools, mock_list_subnets, mock_create_port):
        # faking list_subnetpools
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_kuryr_subnetpool_id2 = uuidutils.generate_uuid()

        fake_name = lib_utils.get_neutron_subnetpool_name(FAKE_IP4_CIDR)
        kuryr_subnetpools = self._get_fake_v4_subnetpools(
            fake_kuryr_subnetpool_id, prefixes=[FAKE_IP4_CIDR],
            name=fake_name)
        mock_list_subnetpools.return_value = kuryr_subnetpools

        # faking list_subnets
        docker_endpoint_id = lib_utils.get_hash()
        neutron_network_id = uuidutils.generate_uuid()
        neutron_network_id2 = uuidutils.generate_uuid()
        neutron_subnet_v4_id = uuidutils.generate_uuid()
        neutron_subnet_v4_id2 = uuidutils.generate_uuid()

        # Fake existing Neutron subnets
        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id, neutron_subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        # Making existing Neutron subnet has no tag attribute
        del fake_v4_subnet['subnet']['tags']

        fake_v4_subnet2 = self._get_fake_v4_subnet(
            neutron_network_id2, docker_endpoint_id, neutron_subnet_v4_id2,
            subnetpool_id=fake_kuryr_subnetpool_id2,
            cidr=FAKE_IP4_CIDR)
        # Making existing Neutron subnet has no tag attribute
        del fake_v4_subnet2['subnet']['tags']

        fake_subnet_response = {
            'subnets': []
        }
        fake_subnet_response2 = {
            'subnets': [
                fake_v4_subnet2['subnet'],
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.side_effect = [
            fake_subnet_response, fake_subnet_response2]

        # Testing container ip allocation
        fake_mac_address = 'fa:16:3e:86:a0:fe'
        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': '',  # Querying for container address
            'Options': {const.DOCKER_MAC_ADDRESS_OPTION: fake_mac_address}
        }
        response = self.app.post('/IpamDriver.RequestAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(w_exceptions.InternalServerError.code,
                         response.status_code)
        mock_list_subnetpools.assert_called_with(
            id=fake_kuryr_subnetpool_id)
        mock_list_subnets.assert_called_with(
            cidr=FAKE_IP4_CIDR)
        decoded_json = jsonutils.loads(response.data)
        self.assertIn('Err', decoded_json)
        self.assertIn(fake_kuryr_subnetpool_id, decoded_json['Err'])

    @mock.patch('kuryr_libnetwork.controllers._neutron_port_add_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((False), (True))
    def test_ipam_driver_request_address_overlapping_cidr_in_kuryr(
            self, use_tag_ext, mock_app, mock_list_subnetpools,
            mock_list_subnets, mock_create_port, mock_port_add_tag):
        mock_app.tag_ext = use_tag_ext
        # faking list_subnetpools
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()

        fake_name = lib_utils.get_neutron_subnetpool_name(FAKE_IP4_CIDR)
        kuryr_subnetpools = self._get_fake_v4_subnetpools(
            fake_kuryr_subnetpool_id, prefixes=[FAKE_IP4_CIDR],
            name=fake_name)
        mock_list_subnetpools.return_value = kuryr_subnetpools

        # faking list_subnets
        docker_endpoint_id = lib_utils.get_hash()
        neutron_network_id = uuidutils.generate_uuid()
        neutron_network_id2 = uuidutils.generate_uuid()
        neutron_subnet_v4_id = uuidutils.generate_uuid()
        neutron_subnet_v4_id2 = uuidutils.generate_uuid()

        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id, neutron_subnet_v4_id,
            cidr=FAKE_IP4_CIDR,
            name=utils.make_subnet_name(FAKE_IP4_CIDR))

        fake_v4_subnet2 = self._get_fake_v4_subnet(
            neutron_network_id2, docker_endpoint_id, neutron_subnet_v4_id2,
            cidr=FAKE_IP4_CIDR,
            name=utils.make_subnet_name(FAKE_IP4_CIDR))

        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet'],
                fake_v4_subnet2['subnet'],
            ]
        }
        mock_list_subnets.return_value = fake_subnet_response
        # faking create_port
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_mac_address = 'fa:16:3e:86:a0:fe'
        fake_port = self._get_fake_port(
            docker_endpoint_id, neutron_network_id,
            fake_neutron_port_id,
            neutron_subnet_v4_id=neutron_subnet_v4_id,
            neutron_subnet_v4_address="10.0.0.5",
            neutron_mac_address=fake_mac_address)
        mock_create_port.return_value = fake_port
        port_request = {
            'name': const.KURYR_UNBOUND_PORT,
            'admin_state_up': True,
            'network_id': neutron_network_id,
        }
        port_request['fixed_ips'] = []
        fixed_ip = {'subnet_id': neutron_subnet_v4_id}
        port_request['fixed_ips'].append(fixed_ip)

        # Testing container ip allocation
        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': '',  # Querying for container address
            'Options': {const.DOCKER_MAC_ADDRESS_OPTION: fake_mac_address}
        }
        mock_port_add_tag.return_value = None
        response = self.app.post('/IpamDriver.RequestAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnetpools.assert_called_with(
            id=fake_kuryr_subnetpool_id)
        mock_list_subnets.assert_called_with(
            subnetpool_id=fake_kuryr_subnetpool_id)
        mock_create_port.assert_called_with(
            {'port': port_request})
        if mock_app.tag_ext:
            mock_port_add_tag.assert_called()
        else:
            mock_port_add_tag.assert_not_called()
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual('10.0.0.5/16', decoded_json['Address'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    def test_ipam_driver_request_address_for_same_gateway(self,
            mock_list_subnets):
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        # faking list_subnets
        docker_endpoint_id = lib_utils.get_hash()
        neutron_network_id = uuidutils.generate_uuid()
        subnet_v4_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_v4_subnet['subnet'].update(gateway_ip='10.0.0.1')
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.return_value = fake_subnet_response

        # Testing container ip allocation
        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': '10.0.0.1',
            'Options': {
                const.REQUEST_ADDRESS_TYPE: const.NETWORK_GATEWAY_OPTIONS
            }
        }
        response = self.app.post('/IpamDriver.RequestAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(
            subnetpool_id=fake_kuryr_subnetpool_id)
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual('10.0.0.1/16', decoded_json['Address'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    def test_ipam_driver_request_address_for_different_gateway(self,
            mock_list_subnets):
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        # faking list_subnets
        docker_endpoint_id = lib_utils.get_hash()
        neutron_network_id = uuidutils.generate_uuid()
        subnet_v4_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            neutron_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_v4_subnet['subnet'].update(gateway_ip='10.0.0.1')
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.return_value = fake_subnet_response

        # Testing container ip allocation
        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': '10.0.0.5',  # Different with existed gw ip
            'Options': {
                const.REQUEST_ADDRESS_TYPE: const.NETWORK_GATEWAY_OPTIONS
            }
        }
        response = self.app.post('/IpamDriver.RequestAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(500, response.status_code)
        mock_list_subnets.assert_called_with(
            subnetpool_id=fake_kuryr_subnetpool_id)
        decoded_json = jsonutils.loads(response.data)
        self.assertIn('Err', decoded_json)
        err_message = ("Requested gateway {0} does not match with "
                       "gateway {1} in existed network.").format(
                           '10.0.0.5', '10.0.0.1')
        self.assertEqual({'Err': err_message}, decoded_json)

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.delete_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    def test_ipam_driver_release_address(self,
            mock_list_subnets, mock_list_ports, mock_delete_port):
        # faking list_subnets
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        docker_network_id = lib_utils.get_hash()
        docker_endpoint_id = lib_utils.get_hash()
        subnet_v4_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            docker_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR,
            name=utils.make_subnet_name(FAKE_IP4_CIDR))
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.return_value = fake_subnet_response

        fake_ip4 = '10.0.0.5'
        # faking list_ports and delete_port
        neutron_network_id = uuidutils.generate_uuid()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_port = base.TestKuryrBase._get_fake_port(
            docker_endpoint_id, neutron_network_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            subnet_v4_id,
            neutron_subnet_v4_address=fake_ip4,
            device_owner=lib_const.DEVICE_OWNER,
            tags=lib_const.DEVICE_OWNER)

        fake_port['port']['fixed_ips'] = [
            {'subnet_id': subnet_v4_id, 'ip_address': fake_ip4}
        ]

        list_port_response = {'ports': [fake_port['port']]}
        mock_list_ports.return_value = list_port_response

        mock_delete_port.return_value = {}

        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': fake_ip4
        }
        response = self.app.post('/IpamDriver.ReleaseAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(
            subnetpool_id=fake_kuryr_subnetpool_id)
        mock_list_ports.assert_called()
        mock_delete_port.assert_called_with(
            fake_port['port']['id'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.delete_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    def test_ipam_driver_release_address_w_existing_subnet(self,
            mock_list_subnetpools, mock_list_subnets, mock_list_ports,
            mock_delete_port):
        # faking list_subnetpools
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_name = str('-'.join(['kuryrPool', FAKE_IP4_CIDR]))
        kuryr_subnetpools = self._get_fake_v4_subnetpools(
            fake_kuryr_subnetpool_id, prefixes=[FAKE_IP4_CIDR], name=fake_name)
        mock_list_subnetpools.return_value = kuryr_subnetpools

        # faking list_subnets
        docker_network_id = lib_utils.get_hash()
        docker_endpoint_id = lib_utils.get_hash()
        subnet_v4_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            docker_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_subnet_response = {
            'subnets': []
        }
        fake_subnet_response2 = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.side_effect = [
            fake_subnet_response, fake_subnet_response2]

        fake_ip4 = '10.0.0.5'
        # faking list_ports and delete_port
        neutron_network_id = uuidutils.generate_uuid()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_port = base.TestKuryrBase._get_fake_port(
            docker_endpoint_id, neutron_network_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            subnet_v4_id,
            neutron_subnet_v4_address=fake_ip4,
            device_owner=lib_const.DEVICE_OWNER,
            tags=lib_const.DEVICE_OWNER)

        fake_port['port']['fixed_ips'] = [
            {'subnet_id': subnet_v4_id, 'ip_address': fake_ip4}
        ]

        list_port_response = {'ports': [fake_port['port']]}
        mock_list_ports.return_value = list_port_response

        mock_delete_port.return_value = {}

        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': fake_ip4
        }
        response = self.app.post('/IpamDriver.ReleaseAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnetpools.assert_called_with(
            id=fake_kuryr_subnetpool_id)
        mock_list_subnets.assert_called_with(
            cidr=FAKE_IP4_CIDR)
        mock_list_ports.assert_called()
        mock_delete_port.assert_called_with(
            fake_port['port']['id'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.remove_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.update_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.delete_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app')
    def test_ipam_driver_release_address_w_existing_port(self,
            mock_app, mock_list_subnets, mock_list_ports, mock_delete_port,
            mock_update_port, mock_remove_tag):
        # TODO(hongbin): Current implementation still delete existing ports
        # if tag extension is not enabled. This needs to be fixed and test
        # case needs to be added after.
        mock_app.tag_ext = True
        # faking list_subnets
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        docker_network_id = lib_utils.get_hash()
        docker_endpoint_id = lib_utils.get_hash()
        subnet_v4_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            docker_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.return_value = fake_subnet_response

        fake_ip4 = '10.0.0.5'
        # faking list_ports and delete_port
        neutron_network_id = uuidutils.generate_uuid()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_port = base.TestKuryrBase._get_fake_port(
            docker_endpoint_id, neutron_network_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            subnet_v4_id,
            neutron_subnet_v4_address=fake_ip4,
            device_owner=lib_const.DEVICE_OWNER,
            tags=const.KURYR_EXISTING_NEUTRON_PORT)

        fake_port['port']['fixed_ips'] = [
            {'subnet_id': subnet_v4_id, 'ip_address': fake_ip4}
        ]

        list_port_response = {'ports': [fake_port['port']]}
        mock_list_ports.return_value = list_port_response

        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': fake_ip4
        }
        response = self.app.post('/IpamDriver.ReleaseAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(
            subnetpool_id=fake_kuryr_subnetpool_id)
        mock_list_ports.assert_called()
        expect_updated_port = {'device_owner': '',
                               'device_id': '', 'binding:host_id': ''}
        mock_update_port.assert_called_with(fake_port['port']['id'],
                                            {'port': expect_updated_port})
        mock_delete_port.assert_not_called()
        mock_remove_tag.assert_called_with('ports',
                                           fake_port['port']['id'],
                                           const.KURYR_EXISTING_NEUTRON_PORT)

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.delete_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app')
    @ddt.data((False), (True))
    def test_ipam_driver_release_address_w_same_ip_and_cidr(
            self, use_tag_ext, mock_app, mock_list_subnets, mock_list_ports,
            mock_delete_port):
        # It checks only the kuryr port is removed even if the other port
        # has the same IP and belongs to a subnet with the same subnetpool_id
        # faking list_subnets
        mock_app.tag_ext = use_tag_ext
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        # faking list_subnets
        docker_network_id = lib_utils.get_hash()
        docker_endpoint_id = lib_utils.get_hash()
        no_kuryr_endpoint_id = lib_utils.get_hash()
        subnet_v4_id = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            docker_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet']
            ]
        }
        mock_list_subnets.return_value = fake_subnet_response

        fake_ip4 = '10.0.0.5'
        # faking list_ports and delete_port
        neutron_network_id = uuidutils.generate_uuid()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_port = base.TestKuryrBase._get_fake_port(
            docker_endpoint_id, neutron_network_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            subnet_v4_id,
            neutron_subnet_v4_address=fake_ip4,
            device_owner=lib_const.DEVICE_OWNER)
        fake_port_no_kuryr = base.TestKuryrBase._get_fake_port(
            no_kuryr_endpoint_id, neutron_network_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            subnet_v4_id,
            neutron_subnet_v4_address=fake_ip4)
        fake_port_no_kuryr['port']['name'] = 'port0'

        if use_tag_ext:
            fake_port['port']['tags'] = [lib_const.DEVICE_OWNER]
        fake_port['port']['fixed_ips'] = [
            {'subnet_id': subnet_v4_id, 'ip_address': fake_ip4}
        ]
        fake_port_no_kuryr['port']['fixed_ips'] = [
            {'subnet_id': subnet_v4_id, 'ip_address': fake_ip4}
        ]

        list_port_response = {'ports': [fake_port['port'],
                                        fake_port_no_kuryr['port']]}
        mock_list_ports.return_value = list_port_response

        mock_delete_port.return_value = {}

        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': fake_ip4
        }
        response = self.app.post('/IpamDriver.ReleaseAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(
            subnetpool_id=fake_kuryr_subnetpool_id)
        mock_list_ports.assert_called()
        mock_delete_port.assert_called_once_with(
            fake_port['port']['id'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.remove_tag')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.update_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.delete_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app')
    def test_ipam_driver_release_address_w_existing_port_w_same_ip_and_cidr(
            self, mock_app, mock_list_subnets, mock_list_ports,
            mock_delete_port, mock_update_port, mock_remove_tag):
        # TODO(hongbin): Current implementation still delete existing ports
        # if tag extension is not enabled. This needs to be fixed and test
        # case needs to be added after.
        mock_app.tag_ext = True
        fake_kuryr_subnetpool_id = uuidutils.generate_uuid()
        fake_kuryr_subnetpool_id2 = uuidutils.generate_uuid()
        # faking list_subnets
        docker_network_id = lib_utils.get_hash()
        docker_network_id2 = lib_utils.get_hash()
        docker_endpoint_id = lib_utils.get_hash()
        docker_endpoint_id2 = lib_utils.get_hash()
        subnet_v4_id = uuidutils.generate_uuid()
        subnet_v4_id2 = uuidutils.generate_uuid()
        fake_v4_subnet = self._get_fake_v4_subnet(
            docker_network_id, docker_endpoint_id, subnet_v4_id,
            subnetpool_id=fake_kuryr_subnetpool_id,
            cidr=FAKE_IP4_CIDR)
        fake_v4_subnet2 = self._get_fake_v4_subnet(
            docker_network_id2, docker_endpoint_id2, subnet_v4_id2,
            subnetpool_id=fake_kuryr_subnetpool_id2,
            cidr=FAKE_IP4_CIDR)
        fake_subnet_response = {
            'subnets': [
                fake_v4_subnet['subnet'], fake_v4_subnet2['subnet']
            ]
        }
        mock_list_subnets.return_value = fake_subnet_response

        fake_ip4 = '10.0.0.5'
        # faking list_ports and delete_port
        neutron_network_id = uuidutils.generate_uuid()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_port = base.TestKuryrBase._get_fake_port(
            docker_endpoint_id, neutron_network_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            subnet_v4_id,
            neutron_subnet_v4_address=fake_ip4,
            device_owner=lib_const.DEVICE_OWNER,
            tags=[const.KURYR_EXISTING_NEUTRON_PORT])
        fake_port2 = base.TestKuryrBase._get_fake_port(
            docker_endpoint_id2, neutron_network_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            subnet_v4_id2,
            neutron_subnet_v4_address=fake_ip4,
            device_owner=lib_const.DEVICE_OWNER,
            tags=[const.KURYR_EXISTING_NEUTRON_PORT])

        fake_port['port']['fixed_ips'] = [
            {'subnet_id': subnet_v4_id, 'ip_address': fake_ip4}
        ]
        fake_port2['port']['fixed_ips'] = [
            {'subnet_id': subnet_v4_id2, 'ip_address': fake_ip4}
        ]

        list_port_response = {'ports': [fake_port['port'],
                                        fake_port2['port']]}
        mock_list_ports.return_value = list_port_response

        mock_delete_port.return_value = {}

        fake_request = {
            'PoolID': fake_kuryr_subnetpool_id,
            'Address': fake_ip4
        }
        response = self.app.post('/IpamDriver.ReleaseAddress',
                                content_type='application/json',
                                data=jsonutils.dumps(fake_request))

        self.assertEqual(200, response.status_code)
        mock_list_subnets.assert_called_with(
            subnetpool_id=fake_kuryr_subnetpool_id)
        mock_list_ports.assert_called()
        expect_updated_port = {'device_owner': '',
                               'device_id': '', 'binding:host_id': ''}
        mock_update_port.assert_called_once_with(fake_port['port']['id'],
                                            {'port': expect_updated_port})
        mock_delete_port.assert_not_called()
        mock_remove_tag.assert_called_once_with('ports',
                                           fake_port['port']['id'],
                                           const.KURYR_EXISTING_NEUTRON_PORT)



# Copyright 2016: IBM Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import docker
import utils

from rally_openstack import scenario


@scenario.configure(name="Kuryr.list_networks")
class KuryrListNetworks(utils.KuryrScenario):

    def run(self, network_list_args=None):
        """List the networks.

        Measure the "docker network ls" command performance under kuryr.

        This will call the docker client API to list networks

        TODO (baohua):
        1. may support tenant/user in future.
        2. validation.required_services add KURYR support

        :param network_list_args: dict: names, ids
        """
        self.docker_client = docker.APIClient(base_url='tcp://0.0.0.0:2375')
        self._list_networks(network_list_args or {})


@scenario.configure(name="Kuryr.create_and_delete_networks_with_kuryr")
class KuryrCreateDeleteNetworksWithKuryr(utils.KuryrScenario):

    def run(self, network_create_args=None):
        """Create and delete a network with kuryr.

        Measure the "docker network create" and "docker network rm" command
        performance with kuryr driver.

        :param network_create_args: dict as options to create the network
        """
        self.docker_client = docker.APIClient(base_url='tcp://0.0.0.0:2375')
        network = self._create_network(is_kuryr=True,
                       network_create_args=network_create_args or {})
        self._delete_network(network)


@scenario.configure(name="Kuryr.create_and_delete_networks_without_kuryr")
class KuryrCreateDeleteNetworksWithoutKuryr(utils.KuryrScenario):

    def run(self, network_create_args=None):
        """Create and delete a network without kuryr.

        Measure the "docker network create" and "docker network rm" command
        performance with default driver.

        :param network_create_args: dict as options to create the network
        """
        self.docker_client = docker.APIClient(base_url='tcp://0.0.0.0:2375')
        network = self._create_network(is_kuryr=False,
                       network_create_args=network_create_args or {})
        self._delete_network(network)


@scenario.configure(name="Kuryr.start_and_stop_containers")
class KuryrStartStopContainers(utils.KuryrScenario):

    def run(self, container_create_args=None):
        """Start and stop container on docker network.

        Measure the "docker run" , "docker stop", "docker rm"
        command performance.
        """
        self.docker_client = docker.APIClient(base_url='tcp://0.0.0.0:2375')
        container_id = self._start_container(container_create_args or {})
        self._stop_container(container_id)
        # TODO(yedongcan) We will hit the Docker bug:
        # "Unable to remove filesystem - device or resource busy"
        # Temporary workaround is disable remove_container here.
        # self._remove_container(container_id)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from kuryr_libnetwork.schemata import commons

REQUEST_POOL_SCHEMA = {
    u'links': [{
        u'method': u'POST',
        u'href': u'/IpamDriver.RequestPool',
        u'description': u'Allocate pool of ip addresses',
        u'rel': u'self',
        u'title': u'Create'
    }],
    u'title': u'Create pool',
    u'required': [u'AddressSpace', u'Pool', u'SubPool', u'V6'],
    u'definitions': {u'commons': {}},
    u'$schema': u'http://json-schema.org/draft-04/hyper-schema',
    u'type': u'object',
    u'properties': {
        u'AddressSpace': {
            u'description': u'The name of the address space.',
            u'type': u'string',
            u'example': u'foo',
        },
        u'Pool': {
            u'description': u'A range of IP Addresses represented in '
                            u'CIDR format address/mask.',
            u'$ref': u'#/definitions/commons/definitions/cidrv4_or_cidrv6'
        },
        u'SubPool': {
            u'description': u'A subset of IP range from Pool in'
                            u'CIDR format address/mask.',
            u'$ref': u'#/definitions/commons/definitions/cidrv4_or_cidrv6'
        },
        u'Options': {
            u'type': [u'object', u'null'],
            u'description': u'Options',
            u'example': {},
        },
        u'V6': {
            u'description': u'If set to "True", requesting IPv6 pool and '
                            u'vice-versa.',
            u'type': u'boolean',
            u'example': False
        }
    }
}

REQUEST_POOL_SCHEMA[u'definitions'][u'commons'] = commons.COMMONS



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

__all__ = [
    'list_kuryr_libnetwork_opts',
]

import copy
import itertools

from oslo_log import _options

from kuryr.lib import opts as lib_opts
from kuryr_libnetwork import config


_core_opts_with_logging = config.core_opts
_core_opts_with_logging += _options.common_cli_opts
_core_opts_with_logging += _options.logging_cli_opts
_core_opts_with_logging += _options.generic_log_opts

_kuryr_libnetwork_opts = [
    (None, list(itertools.chain(_core_opts_with_logging))),
]


def list_kuryr_libnetwork_opts():
    """Return a list of oslo_config options available in kuryr-libnetwork service.

    Each element of the list is a tuple. The first element is the name of the
    group under which the list of elements in the second element will be
    registered. A group name of None corresponds to the [DEFAULT] group in
    config files.

    This function is also discoverable via the 'kuryr_libnetwork' entry point
    under the 'oslo_config.opts' namespace.

    The purpose of this is to allow tools like the Oslo sample config file
    generator to discover the options exposed to users by kuryr-libnetwork.

    :returns: a list of (group_name, opts) tuples
    """

    return ([(k, copy.deepcopy(o)) for k, o in _kuryr_libnetwork_opts] +
            lib_opts.list_kuryr_opts())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from neutronclient.common import exceptions as n_exceptions
from oslo_log import log

from kuryr.lib import binding
from kuryr.lib import exceptions

from kuryr_libnetwork import app
from kuryr_libnetwork.port_driver import base

LOG = log.getLogger(__name__)


class NestedDriver(base.BaseNestedDriver):
    """Driver for container-in-VM deployments with MACVLAN and IPVLAN."""

    BINDING_DRIVERS = ('macvlan', 'ipvlan')

    def __init__(self):
        super(NestedDriver, self).__init__()

    def get_supported_bindings(self):
        """Returns a tuple of supported binding driver names for the driver.

        :returns: a tuple of strings
        """
        return self.BINDING_DRIVERS

    def get_default_network_id(self):
        """Returns a Neutron network ID as per driver logic, if any.

        Nested Endpoints need to join the same network as their Master
        interface, this function will return its Neutron network UUID for the
        Endpoint to join or throw in case of failure.

        :returns: the Neutron network UUID as a string
        :raises: exceptions.KuryrException
        """
        vm_port = self._get_port_from_host_iface(self.link_iface)

        return vm_port['network_id']

    def create_host_iface(self, endpoint_id, neutron_port, subnets,
                          network=None):
        """Instantiates a host interface and binds it to the host.

        A host linked interface will be created for the specific Neutron port
        by delegating to the pre-selected kuryr-lib driver.
        This driver will also add the IP and MAC address pairs of the Endpoint
        to the allowed_address_pairs list of the Neutron port associated to the
        underlying host interface.

        :param endpoint_id:  the ID of the endpoint as string
        :param neutron_port: the container Neutron port dictionary as returned
                             by python-neutronclient
        :param subnets:      an iterable of all the Neutron subnets which the
                             endpoint is trying to join
        :param network:      the Neutron network which the endpoint is trying
                             to join
        :returns: the tuple of stdout and stderr returned by
                  processutils.execute invoked
                  with the executable script for binding
        :raises: exceptions.VethCreationFailure,
                 exceptions.KuryrException,
                 n_exceptions.NeutronClientException,
                 processutils.ProcessExecutionError
        """
        container_mac = neutron_port['mac_address']
        container_ips = neutron_port['fixed_ips']

        if not container_ips:  # The MAC address should be mandatory, no check
            raise exceptions.KuryrException(
                "Neutron port {0} does not have fixed_ips."
                .format(neutron_port['id']))

        vm_port = self._get_port_from_host_iface(self.link_iface)

        _, _, (stdout, stderr) = binding.port_bind(
            endpoint_id, neutron_port, subnets, network, vm_port)
        self._add_to_allowed_address_pairs(vm_port, container_ips,
                                           container_mac)

        return (stdout, stderr)

    def delete_host_iface(self, endpoint_id, neutron_port):
        """Deletes a host interface after unbinding it from the host.

        The host Slave interface associated to the Neutron port will be deleted
        by delegating to the selected kuryr-lib driver.
        This driver will also remove the IP and MAC address pairs of the
        Endpoint to the allowed_address_pairs list of the Neutron port
        associated to the underlying host interface.

        :param endpoint_id:  the ID of the endpoint as string
        :param neutron_port: a port dictionary returned from
                             python-neutronclient
        :returns: the tuple of stdout and stderr returned
                  by processutils.execute invoked with the executable script
                  for unbinding
        :raises: exceptions.VethDeletionFailure,
                 exceptions.KuryrException,
                 n_exceptions.NeutronClientException,
                 processutils.ProcessExecutionError,
        """
        vm_port = self._get_port_from_host_iface(self.link_iface)
        container_ips = neutron_port['fixed_ips']

        self._remove_from_allowed_address_pairs(vm_port, container_ips)
        return binding.port_unbind(endpoint_id, neutron_port)

    def _add_to_allowed_address_pairs(self, port, ip_addresses,
                                      mac_address=None):
        address_pairs = port['allowed_address_pairs']
        for ip_entry in ip_addresses:
            pair = {'ip_address': ip_entry['ip_address']}
            if mac_address:
                pair['mac_address'] = mac_address
            address_pairs.append(pair)

        self._update_port_address_pairs(port['id'], address_pairs)

    def _remove_from_allowed_address_pairs(self, port, ip_addresses):
        address_pairs = port['allowed_address_pairs']
        filter = frozenset(ip_entry['ip_address'] for ip_entry in ip_addresses)
        updated_address_pairs = []

        # filter allowed IPs by copying
        for address_pair in address_pairs:
            if address_pair['ip_address'] in filter:
                continue
            updated_address_pairs.append(address_pair)

        self._update_port_address_pairs(port['id'], updated_address_pairs)

    def _update_port_address_pairs(self, port_id, address_pairs):
        try:
            app.neutron.update_port(
                port_id,
                {
                    'port': {
                        'allowed_address_pairs': address_pairs
                    }
                })
        except n_exceptions.NeutronClientException as ex:
            LOG.error("Error happened during updating Neutron "
                      "port %(port_id)s: %(ex)s", port_id, ex)
            raise



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import os
import sys
from unittest import mock
from urllib import parse

from neutronclient.common import exceptions as n_exceptions

from kuryr.lib import exceptions
from kuryr_libnetwork import config
from kuryr_libnetwork import controllers
from kuryr_libnetwork.server import start
from kuryr_libnetwork.tests.unit import base


@ddt.ddt
class ConfigurationTest(base.TestKuryrBase):

    def test_defaults(self):
        basepath = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                '../../..'))
        self.assertEqual(basepath,
                         config.CONF.pybasedir)
        self.assertEqual('/usr/libexec/kuryr',
                         config.CONF.bindir)
        self.assertEqual('http://127.0.0.1:23750',
                         config.CONF.kuryr_uri)

        self.assertEqual('kuryr',
                         config.CONF.neutron.default_subnetpool_v4)

        self.assertEqual('kuryr6',
                         config.CONF.neutron.default_subnetpool_v6)

        self.assertEqual('kuryr_libnetwork.port_driver.drivers.veth',
                         config.CONF.port_driver)

    @mock.patch('kuryr_libnetwork.controllers.check_for_neutron_tag_support')
    @mock.patch('kuryr_libnetwork.controllers.check_for_neutron_ext_support')
    @mock.patch('kuryr_libnetwork.controllers.neutron_client')
    @mock.patch('kuryr_libnetwork.app.run')
    def test_start(self, mock_run, mock_neutron_client,
                   mock_check_neutron_ext_support,
                   mock_check_for_neutron_tag_support):
        with mock.patch.object(sys, 'argv', ['prog']):
            start()
        kuryr_uri = parse.urlparse(config.CONF.kuryr_uri)
        mock_neutron_client.assert_called_once()
        mock_check_neutron_ext_support.assert_called_once()
        mock_check_for_neutron_tag_support.assert_called_once_with()
        mock_run.assert_called_once_with(kuryr_uri.hostname, 23750,
            ssl_context=None)

    def test_check_for_neutron_ext_support_with_ex(self):
        with mock.patch.object(controllers.app.neutron,
                            'show_extension') as mock_extension:
            ext_alias = "subnet_allocation"
            err = n_exceptions.NotFound.status_code
            ext_not_found_ex = n_exceptions.NeutronClientException(
                status_code=err,
                message="")
            mock_extension.side_effect = ext_not_found_ex
            ex = exceptions.MandatoryApiMissing
            self.assertRaises(ex, controllers.check_for_neutron_ext_support)
            mock_extension.assert_called_once_with(ext_alias)

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.show_extension')
    def test_check_for_neutron_tag_support_with_modern_ext(self,
                                                           mock_extension):
        controllers.check_for_neutron_tag_support()
        mock_extension.assert_called_once_with('standard-attr-tag')
        self.assertTrue(controllers.app.tag)
        self.assertTrue(controllers.app.tag_ext)

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.show_extension')
    @ddt.data({'tag': True, 'tag-ext': True},
              {'tag': True, 'tag-ext': False},
              {'tag': False, 'tag-ext': True},
              {'tag': False, 'tag-ext': False})
    def test_check_for_neutron_tag_support_with_legacy_ext(self, ext_matrix,
                                                           mock_extension):
        err = n_exceptions.NotFound.status_code
        ext_not_found_ex = n_exceptions.NeutronClientException(
            status_code=err,
            message="")

        def mock_fn(ext):
            if not ext_matrix.get(ext, False):
                raise ext_not_found_ex

        mock_extension.side_effect = mock_fn
        controllers.check_for_neutron_tag_support()
        mock_extension.assert_any_call('standard-attr-tag')
        mock_extension.assert_any_call('tag')
        mock_extension.assert_any_call('tag-ext')

        self.assertEqual(controllers.app.tag, ext_matrix['tag'])
        self.assertEqual(controllers.app.tag_ext, ext_matrix['tag-ext'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import docker

from rally.common import logging
from rally import consts
from rally.task import context

LOG = logging.getLogger(__name__)


@context.configure(name="docker_network", order=1000)
class DockerNetworkContext(context.Context):
    """Create a kuryr or non-kuryr docker network as context"""

    CONFIG_SCHEMA = {
        "type": "object",
        "$schema": consts.JSON_SCHEMA,
        "additionalProperties": False,
        "properties": {
            "is_kuryr": {
                "type": "boolean"
            },
            "Subnet": {
                "type": "string"
            },
            "IPRange": {
                "type": "string"
            },
            "Gateway": {
                "type": "string"
            }
        }
    }
    DEFAULT_CONFIG = {
        "is_kuryr": True,
        "Subnet": "50.0.0.0/24",
        "IPRange": "50.0.0.0/24",
        "Gateway": "50.0.0.1"
    }

    def setup(self):
        """Create kuryr or non-kuryr docker network, and prepare image cache"""
        try:
            docker_client = docker.APIClient(base_url="tcp://0.0.0.0:2375")

            if self.config["is_kuryr"]:
                ipam = {
                    "Driver": "kuryr",
                    "Options": {},
                    "Config": [
                        {
                            "Subnet": self.config.get("Subnet"),
                            "IPRange": self.config.get("IPRange"),
                            "Gateway": self.config.get("Gateway")
                        }
                    ]
                }
                res = docker_client.create_network(name="kuryr_network",
                                                   driver="kuryr",
                                                   ipam=ipam)
                self.context["netid"] = res.get("Id")
                self.context["netname"] = "kuryr_network"
            else:
                res = docker_client.create_network(name="docker_network")
                self.context["netid"] = res.get("Id")
                self.context["netname"] = "docker_network"
            LOG.debug("Container network id is '%s'" % self.context["netid"])
        except Exception as e:
            msg = "Can't create docker network: %s" % e.message
            if logging.is_debug():
                LOG.exception(msg)
            else:
                LOG.warning(msg)

    def cleanup(self):
        """Clean up network"""
        try:
            self.docker_client.remove_network(self.context["netid"])
            LOG.debug("Docker network '%s' deleted" % self.context["netid"])
        except Exception as e:
            msg = "Can't delete docker network: %s" % e.message
            if logging.is_debug():
                LOG.exception(msg)
            else:
                LOG.warning(msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Routines for configuring Kuryr
"""

import os

from oslo_config import cfg
from oslo_log import log
import pbr.version

from kuryr.lib._i18n import _
from kuryr.lib import config as lib_config

core_opts = [
    cfg.StrOpt('pybasedir',
               default=os.path.abspath(os.path.join(os.path.dirname(__file__),
                                                    '../')),
               help=_('Directory where Kuryr python module is installed.')),
    cfg.StrOpt('kuryr_uri',
               default=os.environ.get('OS_KURYR_URI',
                                      'http://127.0.0.1:23750'),
               help=_('Kuryr URL for accessing Kuryr through json rpc.')),
    cfg.StrOpt('capability_scope',
               default=os.environ.get('CAPABILITY_SCOPE', 'local'),
               choices=['local', 'global'],
               help=_('Kuryr plugin scope reported to libnetwork.')),
    cfg.StrOpt('local_default_address_space',
               default='no_address_space',
               help=_('There is no address-space by default in neutron')),
    cfg.StrOpt('global_default_address_space',
               default='no_address_space',
               help=_('There is no address-space by default in neutron')),
    cfg.StrOpt('port_driver',
               default='kuryr_libnetwork.port_driver.drivers.veth',
               deprecated_for_removal=True,
               help=_('Default driver for the desired deployment model')),
    cfg.StrOpt('default_port_driver',
               default='kuryr_libnetwork.port_driver.drivers.veth',
               help=_('Default driver for the desired deployment model')),
    cfg.ListOpt('enabled_port_drivers',
                default=['kuryr_libnetwork.port_driver.drivers.veth'],
                help=_('Available port drivers')),
    cfg.BoolOpt('process_external_connectivity',
                default=True,
                help=_('Do processing external connectivity')),
    cfg.StrOpt('ssl_cert_file',
               default='/var/lib/kuryr/certs/cert.pem',
               help=_('This option allows setting absolute path'
                    'to the SSL certificate')),
    cfg.StrOpt('ssl_key_file',
               default='/var/lib/kuryr/certs/key.pem',
               help=_('This option allows setting absolute path'
                     'to the SSL private key')),
    cfg.BoolOpt('enable_ssl',
                default=False,
                help=_('Enable SSL for Kuryr'))
]

CONF = cfg.CONF
CONF.register_opts(core_opts)

CONF.register_opts(lib_config.core_opts)
CONF.register_opts(lib_config.binding_opts, 'binding')
lib_config.register_neutron_opts(CONF)

# Setting oslo.log options for logging.
log.register_options(CONF)


def init(args, **kwargs):
    cfg.CONF(
        args=args,
        project='kuryr',
        version=pbr.version.VersionInfo('kuryr-libnetwork').version_string(),
        **kwargs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import ddt
from neutronclient.common import exceptions
from oslo_serialization import jsonutils

from kuryr.lib import utils as lib_utils
from kuryr_libnetwork.tests.unit import base


@ddt.ddt
class TestIpamRequestPoolFailures(base.TestKuryrFailures):
    """Unit tests for testing request pool failures.

    This test covers error responses listed in the spec:
    https://docs.openstack.org/api-ref/network/v2/index.html#subnet-pools-extension-subnetpools # noqa
    """
    def _invoke_create_request(self, pool):
        fake_request = {
            'AddressSpace': '',
            'Pool': pool,
            'SubPool': '',  # In the case --ip-range is not given
            'Options': {},
            'V6': False
        }
        response = self.app.post('/IpamDriver.RequestPool',
                                 content_type='application/json',
                                 data=jsonutils.dumps(fake_request))
        return response

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.create_subnetpool')
    @ddt.data(exceptions.Unauthorized, exceptions.Forbidden,
              exceptions.NotFound)
    def test_request_pool_create_failures(self, GivenException,
            mock_create_subnetpool, mock_list_subnetpools, mock_list_subnets):
        pool_name = lib_utils.get_neutron_subnetpool_name("10.0.0.0/16")
        new_subnetpool = {
            'name': pool_name,
            'default_prefixlen': 16,
            'prefixes': ['10.0.0.0/16'],
            'shared': False}

        fake_subnet = {"subnets": []}
        mock_list_subnets.return_value = fake_subnet

        fake_subnet_pools = {'subnetpools': []}
        mock_list_subnetpools.return_value = fake_subnet_pools

        mock_create_subnetpool.side_effect = GivenException
        pool = '10.0.0.0/16'
        response = self._invoke_create_request(pool)

        self.assertEqual(GivenException.status_code, response.status_code)
        mock_list_subnets.assert_called_with(cidr='10.0.0.0/16')
        mock_list_subnetpools.assert_called_with(name=pool_name)
        mock_create_subnetpool.assert_called_with(
            {'subnetpool': new_subnetpool})
        decoded_json = jsonutils.loads(response.data)
        self.assertIn('Err', decoded_json)
        self.assertEqual(
            {'Err': GivenException.message}, decoded_json)

    def test_request_pool_bad_request_failure(self):
        pool = 'pool-should-be-cidr'
        response = self._invoke_create_request(pool)

        self.assertEqual(400, response.status_code)
        decoded_json = jsonutils.loads(response.data)
        self.assertIn('Err', decoded_json)
        self.assertIn(pool, decoded_json['Err'])
        self.assertIn('Pool', decoded_json['Err'])

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnets')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_subnetpools')
    def test_request_pool_list_subnetpool_failure(self,
                        mock_list_subnetpools, mock_list_subnets):
        fake_subnet = {"subnets": []}
        fake_pool_name = lib_utils.get_neutron_subnetpool_name("10.0.0.0/16")
        mock_list_subnets.return_value = fake_subnet

        ex = exceptions.Unauthorized
        mock_list_subnetpools.side_effect = ex

        pool = '10.0.0.0/16'
        response = self._invoke_create_request(pool)
        mock_list_subnets.assert_called_with(cidr='10.0.0.0/16')
        mock_list_subnetpools.assert_called_with(name=fake_pool_name)
        self.assertEqual(ex.status_code, response.status_code)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from oslo_utils import uuidutils

from kuryr.lib import binding
from kuryr.lib.binding.drivers import utils
from kuryr.lib import constants as lib_const
from kuryr.lib import exceptions
from kuryr.lib import utils as lib_utils
from kuryr_libnetwork.port_driver.drivers import nested
from kuryr_libnetwork.tests.unit import base

mock_interface = mock.MagicMock()


class TestNestedDriver(base.TestKuryrBase):
    """Unit tests for the NestedDriver port driver"""

    def test_get_supported_bindings(self):
        nested_driver = nested.NestedDriver()
        bindings = nested_driver.get_supported_bindings()
        self.assertEqual(bindings, nested.NestedDriver.BINDING_DRIVERS)

    @mock.patch('kuryr_libnetwork.config.CONF')
    @mock.patch.object(nested.NestedDriver, '_get_port_from_host_iface')
    def test_get_default_network_id(self, mock_get_port_from_host, mock_conf):
        mock_conf.binding.link_iface = 'eth0'

        fake_endpoint_id = lib_utils.get_hash()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_neutron_net_id = uuidutils.generate_uuid()
        fake_neutron_v4_subnet_id = uuidutils.generate_uuid()
        fake_neutron_v6_subnet_id = uuidutils.generate_uuid()

        fake_vm_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id)['port']
        mock_get_port_from_host.return_value = fake_vm_port

        nested_driver = nested.NestedDriver()
        host_network_id = nested_driver.get_default_network_id()
        mock_get_port_from_host.assert_called_with('eth0')
        self.assertEqual(host_network_id, fake_vm_port['network_id'])

    @mock.patch('kuryr_libnetwork.config.CONF')
    @mock.patch.object(binding, 'port_bind')
    @mock.patch('kuryr_libnetwork.app.neutron.update_port')
    @mock.patch.object(nested.NestedDriver, '_get_port_from_host_iface')
    def test_create_host_iface(
            self, mock_get_port_from_host,
            mock_update_port, mock_port_bind, mock_conf):
        mock_conf.binding.link_iface = 'eth0'

        fake_endpoint_id = lib_utils.get_hash()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_neutron_net_id = uuidutils.generate_uuid()
        fake_neutron_v4_subnet_id = uuidutils.generate_uuid()
        fake_neutron_v6_subnet_id = uuidutils.generate_uuid()
        fake_vm_port_id = uuidutils.generate_uuid()

        fake_neutron_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            '192.168.1.3', 'fe80::f816:3eff:fe1c:36a9')['port']
        fake_neutron_port['mac_address'] = 'fa:16:3e:20:57:c3'
        fake_vm_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id,
            fake_vm_port_id, lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            '192.168.1.2', 'fe80::f816:3eff:fe20:57c4')['port']
        fake_vm_port['allowed_address_pairs'] = [
            {'ip_address': '192.168.1.2',
             'mac_address': fake_vm_port['mac_address']},
            {'ip_address': 'fe80::f816:3eff:fe20:57c4',
             'mac_address': fake_vm_port['mac_address']}]
        updated_allowed_pairs = fake_vm_port['allowed_address_pairs']
        updated_allowed_pairs.extend([
            {'ip_address': '192.168.1.3',
             'mac_address': fake_neutron_port['mac_address']},
            {'ip_address': 'fe80::f816:3eff:fe1c:36a9',
             'mac_address': fake_neutron_port['mac_address']}])

        fake_subnets = self._get_fake_subnets(
            fake_endpoint_id, fake_neutron_net_id,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id)['subnets']

        fake_network = mock.sentinel.binding_network
        fake_exec_response = ('fake_stdout', '')
        mock_port_bind.return_value = ('fake_host_ifname',
            'fake_container_ifname', fake_exec_response)
        mock_get_port_from_host.return_value = fake_vm_port

        nested_driver = nested.NestedDriver()
        response = nested_driver.create_host_iface(fake_endpoint_id,
            fake_neutron_port, fake_subnets, fake_network)

        mock_get_port_from_host.assert_called_with('eth0')
        mock_port_bind.assert_called_with(fake_endpoint_id,
            fake_neutron_port, fake_subnets, fake_network, fake_vm_port)
        mock_update_port.assert_called_with(
            fake_vm_port['id'],
            {'port': {
                'allowed_address_pairs': updated_allowed_pairs
            }})

        self.assertEqual(response, fake_exec_response)

    @mock.patch('kuryr_libnetwork.config.CONF')
    @mock.patch.object(binding, 'port_unbind')
    @mock.patch('kuryr_libnetwork.app.neutron.update_port')
    @mock.patch.object(nested.NestedDriver, '_get_port_from_host_iface')
    def test_delete_host_iface(
            self, mock_get_port_from_host,
            mock_update_port, mock_port_unbind, mock_conf):
        mock_conf.binding.link_iface = 'eth0'

        fake_endpoint_id = lib_utils.get_hash()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_neutron_net_id = uuidutils.generate_uuid()
        fake_neutron_v4_subnet_id = uuidutils.generate_uuid()
        fake_neutron_v6_subnet_id = uuidutils.generate_uuid()
        fake_vm_port_id = uuidutils.generate_uuid()

        fake_neutron_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id, fake_neutron_port_id,
            lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            '192.168.1.3', 'fe80::f816:3eff:fe1c:36a9')['port']
        fake_neutron_port['mac_address'] = 'fa:16:3e:20:57:c3'
        fake_vm_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id,
            fake_vm_port_id, lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id,
            '192.168.1.2', 'fe80::f816:3eff:fe20:57c4')['port']
        fake_vm_port['allowed_address_pairs'] = [
            {'ip_address': '192.168.1.3',
             'mac_address': fake_neutron_port['mac_address']},
            {'ip_address': 'fe80::f816:3eff:fe1c:36a9',
             'mac_address': fake_neutron_port['mac_address']}]
        updated_allowed_pairs = [
            {'ip_address': '192.168.1.2',
             'mac_address': fake_vm_port['mac_address']},
            {'ip_address': 'fe80::f816:3eff:fe20:57c4',
             'mac_address': fake_vm_port['mac_address']}]
        fake_vm_port['allowed_address_pairs'].extend(updated_allowed_pairs)

        fake_unbind_response = ('fake_stdout', '')
        mock_get_port_from_host.return_value = fake_vm_port
        mock_port_unbind.return_value = fake_unbind_response

        nested_driver = nested.NestedDriver()
        response = nested_driver.delete_host_iface(fake_endpoint_id,
                                                   fake_neutron_port)

        mock_get_port_from_host.assert_called_with('eth0')
        mock_update_port.assert_called_with(
            fake_vm_port['id'],
            {'port': {
                'allowed_address_pairs': updated_allowed_pairs
            }})
        mock_port_unbind.assert_called_with(fake_endpoint_id,
                                            fake_neutron_port)

        self.assertEqual(response, fake_unbind_response)

    @mock.patch.object(utils, 'get_veth_pair_names',
        return_value=("fake_host_ifname", "fake_container_name"))
    def test_get_container_iface_name(self, mock_get_pair_names):
        nested_driver = nested.NestedDriver()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_neutron_port = self._get_fake_port(
            uuidutils.generate_uuid(), uuidutils.generate_uuid(),
            fake_neutron_port_id)['port']
        response = nested_driver.get_container_iface_name(fake_neutron_port)
        mock_get_pair_names.assert_called_with(fake_neutron_port_id)
        self.assertEqual(response, "fake_container_name")


class TestNestedDriverFailures(base.TestKuryrFailures):
    """Unit tests for the NestedDriver port driver failures"""

    @mock.patch.object(nested.NestedDriver, '_get_port_from_host_iface')
    def test_create_host_iface(self, mock_get_port_from_host):
        fake_endpoint_id = lib_utils.get_hash()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_neutron_net_id = uuidutils.generate_uuid()

        fake_neutron_port = self._get_fake_port(
            fake_endpoint_id, fake_neutron_net_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE)['port']

        nested_driver = nested.NestedDriver()
        self.assertRaises(exceptions.KuryrException,
            nested_driver.create_host_iface, fake_endpoint_id,
            fake_neutron_port, None)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os
import sys

from oslo_log import log
from urllib import parse

from kuryr.lib._i18n import _
from kuryr_libnetwork import app
from kuryr_libnetwork import config
from kuryr_libnetwork import controllers


def configure_app():
    config.init(sys.argv[1:])
    log.setup(config.CONF, 'kuryr')
    controllers.neutron_client()
    controllers.check_for_neutron_ext_support()
    controllers.check_for_neutron_tag_support()
    controllers.load_default_subnet_pools()
    controllers.load_port_driver()


def _get_ssl_configs(use_ssl):
    if use_ssl:
        cert_file = config.CONF.ssl_cert_file
        key_file = config.CONF.ssl_key_file

        if not os.path.exists(cert_file):
            raise RuntimeError(
                _("Unable to find cert_file : %s") % cert_file)

        if not os.path.exists(key_file):
            raise RuntimeError(
                _("Unable to find key_file : %s") % key_file)

        return cert_file, key_file
    else:
        return None


def start():
    configure_app()
    kuryr_uri = parse.urlparse(config.CONF.kuryr_uri)

    # SSL configuration
    use_ssl = config.CONF.enable_ssl

    app.run(kuryr_uri.hostname, kuryr_uri.port,
        ssl_context=_get_ssl_configs(use_ssl))


if __name__ == '__main__':
    sys.exit(start())
elif 'UWSGI_ORIGINAL_PROC_NAME' in os.environ:
    # The module is being loaded by uWSGI to get the Flask app running under
    # it. This allows Neutron to be set, since uWSGI does not run 'start',
    # which would trigger the embedded Flask wsgi development server.
    configure_app()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import defaultdict
from itertools import groupby
from operator import itemgetter
from unittest import mock

import ddt
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from kuryr.lib import constants as lib_const
from kuryr.lib import utils as lib_utils
from kuryr_libnetwork import config
from kuryr_libnetwork import constants
from kuryr_libnetwork.tests.unit import base
from kuryr_libnetwork import utils

PORT = 77
SINGLE_PORT = 100
PROTOCOL_TCP = 6
PROTOCOL_UDP = 17


@ddt.ddt
class TestExternalConnectivityKuryr(base.TestKuryrBase):
    """The unitests for external connectivity

    This test class is used to test programming and revoking external
    connectivity for containers.  Tests cover containers which already have
    a security group (perhaps the default security group set up by Neutron)
    associated with their ports in addition to those without any such
    security groups. Tests also cover adding more than one port to the
    list of exposed ports.
    """
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.update_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.show_port')
    @mock.patch(
        'kuryr_libnetwork.controllers.app.neutron.create_security_group_rule')
    @mock.patch(
        'kuryr_libnetwork.controllers.app.neutron.create_security_group')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @ddt.data((False, 1), (True, 1), (False, 2), (True, 2))
    @ddt.unpack
    def test_network_driver_program_external_connectivity(self, existing_sg,
            num_ports, mock_list_ports, mock_create_security_group,
            mock_create_security_group_rule, mock_show_port,
            mock_update_port):
        config.CONF.set_override('process_external_connectivity', True)
        fake_docker_net_id = lib_utils.get_hash()
        fake_docker_endpoint_id = lib_utils.get_hash()

        fake_neutron_net_id = uuidutils.generate_uuid()
        fake_neutron_port_id = uuidutils.generate_uuid()
        neutron_port_name = utils.get_neutron_port_name(
            fake_docker_endpoint_id)
        fake_neutron_v4_subnet_id = uuidutils.generate_uuid()
        fake_neutron_v6_subnet_id = uuidutils.generate_uuid()
        fake_neutron_ports_response = self._get_fake_ports(
            fake_docker_endpoint_id, fake_neutron_net_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id)
        if existing_sg:
            fake_neutron_existing_sec_group_id = uuidutils.generate_uuid()
            fake_neutron_ports_response['ports'][0]['security_groups'] = [
                fake_neutron_existing_sec_group_id]

        mock_list_ports.return_value = fake_neutron_ports_response

        sec_group = {
            'name': utils.get_sg_expose_name(fake_neutron_port_id),
            'description': 'Docker exposed ports created by Kuryr.'
        }
        fake_neutron_sec_group_id = lib_utils.get_hash()
        fake_neutron_sec_group_response = {'security_group':
                                           {'id': fake_neutron_sec_group_id}}
        mock_create_security_group.return_value = (
            fake_neutron_sec_group_response)

        proto_port_dict = defaultdict(list)
        for i in range(num_ports):
            proto_port_dict[constants.PROTOCOLS[PROTOCOL_TCP]].append(PORT + i)
            proto_port_dict[constants.PROTOCOLS[PROTOCOL_UDP]].append(PORT + i)
        proto_port_dict[constants.PROTOCOLS[PROTOCOL_UDP]].append(SINGLE_PORT)

        for proto, port_list in proto_port_dict.items():
            for key, group in groupby(enumerate(sorted(port_list)),
                                      lambda ix: ix[0] - ix[1]):
                port_range_list = list(map(itemgetter(1), group))

                port_range_min = min(port_range_list)
                port_range_max = max(port_range_list)
                sec_group_rule = {
                    'security_group_id': fake_neutron_sec_group_id,
                    'direction': 'ingress',
                    'port_range_min': port_range_min,
                    'port_range_max': port_range_max,
                    'protocol': proto
                }

        sgs = [fake_neutron_sec_group_id]
        if existing_sg:
            sgs.append(fake_neutron_existing_sec_group_id)
        mock_show_port.return_value = {'port':
            fake_neutron_ports_response['ports'][0]}

        port_opt = []
        for i in range(num_ports):
            port_opt.append({u'Port': PORT + i, u'Proto': PROTOCOL_TCP})
            port_opt.append({u'Port': PORT + i, u'Proto': PROTOCOL_UDP})
        port_opt.append({u'Port': SINGLE_PORT, u'Proto': PROTOCOL_UDP})
        options = {'com.docker.network.endpoint.exposedports':
                   port_opt,
                   'com.docker.network.portmap':
                   []}
        data = {
            'NetworkID': fake_docker_net_id,
            'EndpointID': fake_docker_endpoint_id,
            'Options': options,
        }
        response = self.app.post('/NetworkDriver.ProgramExternalConnectivity',
                                 content_type='application/json',
                                 data=jsonutils.dumps(data))

        self.assertEqual(200, response.status_code)
        mock_update_port.assert_called_with(fake_neutron_port_id,
                                {'port': {'security_groups': sgs}})
        mock_show_port.assert_called_with(fake_neutron_port_id)
        mock_create_security_group_rule.assert_called_with(
            {'security_group_rule': sec_group_rule})
        mock_create_security_group.assert_called_with(
            {'security_group': sec_group})
        mock_list_ports.assert_called_with(name=neutron_port_name)
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(constants.SCHEMA['SUCCESS'], decoded_json)

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.update_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.show_port')
    @mock.patch(
        'kuryr_libnetwork.controllers.app.neutron.create_security_group_rule')
    @mock.patch(
        'kuryr_libnetwork.controllers.app.neutron.create_security_group')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @ddt.data((False, 1), (True, 1), (False, 2), (True, 2))
    @ddt.unpack
    def test_network_driver_program_external_connectivity_disabled(
            self, existing_sg,
            num_ports, mock_list_ports, mock_create_security_group,
            mock_create_security_group_rule, mock_show_port,
            mock_update_port):
        config.CONF.set_override('process_external_connectivity', False)
        fake_docker_net_id = lib_utils.get_hash()
        fake_docker_endpoint_id = lib_utils.get_hash()

        port_opt = []
        for i in range(num_ports):
            port_opt.append({u'Port': PORT + i, u'Proto': PROTOCOL_TCP})
            port_opt.append({u'Port': PORT + i, u'Proto': PROTOCOL_UDP})
        port_opt.append({u'Port': SINGLE_PORT, u'Proto': PROTOCOL_UDP})
        options = {'com.docker.network.endpoint.exposedports':
                   port_opt,
                   'com.docker.network.portmap':
                   []}
        data = {
            'NetworkID': fake_docker_net_id,
            'EndpointID': fake_docker_endpoint_id,
            'Options': options,
        }
        response = self.app.post('/NetworkDriver.ProgramExternalConnectivity',
                                 content_type='application/json',
                                 data=jsonutils.dumps(data))

        self.assertEqual(200, response.status_code)
        mock_update_port.assert_not_called()
        mock_show_port.assert_not_called()
        mock_create_security_group_rule.assert_not_called()
        mock_create_security_group.assert_not_called()
        mock_list_ports.assert_not_called()
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(constants.SCHEMA['SUCCESS'], decoded_json)

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.update_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.show_port')
    @mock.patch(
        'kuryr_libnetwork.controllers.app.neutron.delete_security_group')
    @mock.patch(
        'kuryr_libnetwork.controllers.app.neutron.list_security_groups')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @ddt.data((False, False), (False, True), (True, False), (True, True))
    @ddt.unpack
    def test_network_driver_revoke_external_connectivity(self, existing_sg,
            removing_sg, mock_list_ports, mock_list_security_groups,
            mock_delete_security_groups, mock_show_port,
            mock_update_port):
        config.CONF.set_override('process_external_connectivity', True)
        fake_docker_net_id = lib_utils.get_hash()
        fake_docker_endpoint_id = lib_utils.get_hash()

        fake_neutron_net_id = uuidutils.generate_uuid()
        fake_neutron_port_id = uuidutils.generate_uuid()
        fake_neutron_sec_group_id = lib_utils.get_hash()
        neutron_port_name = utils.get_neutron_port_name(
            fake_docker_endpoint_id)
        fake_neutron_v4_subnet_id = uuidutils.generate_uuid()
        fake_neutron_v6_subnet_id = uuidutils.generate_uuid()
        fake_neutron_ports_response = self._get_fake_ports(
            fake_docker_endpoint_id, fake_neutron_net_id,
            fake_neutron_port_id, lib_const.PORT_STATUS_ACTIVE,
            fake_neutron_v4_subnet_id, fake_neutron_v6_subnet_id)
        if existing_sg:
            fake_neutron_existing_sec_group_id = uuidutils.generate_uuid()
            fake_neutron_ports_response['ports'][0]['security_groups'] = [
                fake_neutron_sec_group_id, fake_neutron_existing_sec_group_id]
        else:
            fake_neutron_ports_response['ports'][0]['security_groups'] = [
                fake_neutron_sec_group_id]
        if removing_sg:
            fake_neutron_sec_group_response = {
                'security_groups': [{'id': fake_neutron_sec_group_id}]}
        else:
            fake_neutron_sec_group_response = {
                'security_groups': []}

        mock_list_ports.return_value = fake_neutron_ports_response
        mock_list_security_groups.return_value = (
            fake_neutron_sec_group_response)
        mock_show_port.return_value = {'port':
            fake_neutron_ports_response['ports'][0]}

        if existing_sg:
            sgs = [fake_neutron_existing_sec_group_id]
        else:
            sgs = []
        data = {
            'NetworkID': fake_docker_net_id,
            'EndpointID': fake_docker_endpoint_id,
        }
        response = self.app.post('/NetworkDriver.RevokeExternalConnectivity',
                                 content_type='application/json',
                                 data=jsonutils.dumps(data))

        self.assertEqual(200, response.status_code)
        mock_list_ports.assert_called_with(name=neutron_port_name)
        if removing_sg:
            mock_list_security_groups.assert_called_with(
                name=utils.get_sg_expose_name(fake_neutron_port_id))
            mock_delete_security_groups.assert_called_with(
                fake_neutron_sec_group_id)
            mock_show_port.assert_called_with(fake_neutron_port_id)
            mock_update_port.assert_called_with(fake_neutron_port_id,
                            {'port': {'security_groups': sgs}})
        else:
            mock_list_security_groups.assert_called_with(
                name=utils.get_sg_expose_name(fake_neutron_port_id))
            mock_delete_security_groups.assert_not_called()
            mock_show_port.assert_not_called()
            mock_update_port.assert_not_called()
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(constants.SCHEMA['SUCCESS'], decoded_json)

    @mock.patch('kuryr_libnetwork.controllers.app.neutron.update_port')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.show_port')
    @mock.patch(
        'kuryr_libnetwork.controllers.app.neutron.delete_security_group')
    @mock.patch(
        'kuryr_libnetwork.controllers.app.neutron.list_security_groups')
    @mock.patch('kuryr_libnetwork.controllers.app.neutron.list_ports')
    @ddt.data((False, False), (False, True), (True, False), (True, True))
    @ddt.unpack
    def test_network_driver_revoke_external_connectivity_disabled(
            self, existing_sg,
            removing_sg, mock_list_ports, mock_list_security_groups,
            mock_delete_security_groups, mock_show_port,
            mock_update_port):
        config.CONF.set_override('process_external_connectivity', False)
        fake_docker_net_id = lib_utils.get_hash()
        fake_docker_endpoint_id = lib_utils.get_hash()

        data = {
            'NetworkID': fake_docker_net_id,
            'EndpointID': fake_docker_endpoint_id,
        }
        response = self.app.post('/NetworkDriver.RevokeExternalConnectivity',
                                 content_type='application/json',
                                 data=jsonutils.dumps(data))

        self.assertEqual(200, response.status_code)
        mock_list_ports.assert_not_called()
        mock_list_security_groups.assert_not_called()
        mock_delete_security_groups.assert_not_called()
        mock_show_port.assert_not_called()
        mock_update_port.assert_not_called()
        decoded_json = jsonutils.loads(response.data)
        self.assertEqual(constants.SCHEMA['SUCCESS'], decoded_json)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import glob
import os
import re

from kuryr.lib import binding
from kuryr.lib import exceptions

from kuryr_libnetwork import constants as const
from kuryr_libnetwork.port_driver import driver


def get_ifname_by_pci_address(pci_addr, pf_interface=False):
    """Get the interface name based on a VF's pci address.

    The returned interface name is either the parent PF's or that of the VF
    itself based on the argument of pf_interface.
    """
    dev_path = _get_sysfs_netdev_path(pci_addr, pf_interface)
    try:
        dev_info = os.listdir(dev_path)
        return dev_info.pop()
    except Exception:
        raise exceptions.KuryrException(
            "PCI device %s not found" % pci_addr)


def _get_sysfs_netdev_path(pci_addr, pf_interface):
    """Get the sysfs path based on the PCI address of the device.

    Assumes a networking device - will not check for the existence of the path.
    """
    if pf_interface:
        return "/sys/bus/pci/devices/%s/physfn/net" % pci_addr
    return "/sys/bus/pci/devices/%s/net" % pci_addr


def get_vf_num_by_pci_address(pci_addr):
    """Get the VF number based on a VF's pci address

    A VF is associated with an VF number, which ip link command uses to
    configure it. This number can be obtained from the PCI device filesystem.
    """
    VIRTFN_RE = re.compile(r"virtfn(\d+)")
    virtfns_path = "/sys/bus/pci/devices/%s/physfn/virtfn*" % (pci_addr)
    vf_num = None
    try:
        for vf_path in glob.iglob(virtfns_path):
            if re.search(pci_addr, os.readlink(vf_path)):
                t = VIRTFN_RE.search(vf_path)
                vf_num = t.group(1)
                break
    except Exception:
        pass
    if vf_num is None:
        raise exceptions.KuryrException(
            "PCI device %s not found" % pci_addr)
    return vf_num


class SriovDriver(driver.Driver):
    """Driver supporting SR-IOV on Bare Metal"""

    BINDING_DRIVERS = ('hw_veb',)

    def get_supported_bindings(self):
        """Returns a tuple of supported binding driver names for the driver.

        :returns: a tuple of strings
        """
        return self.BINDING_DRIVERS

    def get_default_network_id(self):
        """Returns a Neutron network ID as per driver logic, if any.

        This driver does not make use of any specific network and will thus
        return None.

        :returns: None
        """
        return None

    def create_host_iface(self, endpoint_id, neutron_port, subnets,
                          network=None):
        """Instantiates a host interface and bind it to the host.

        :param endpoint_id:  the ID of the endpoint as string
        :param neutron_port: the container Neutron port dictionary as returned
                             by python-neutronclient
        :param subnets:      an iterable of all the Neutron subnets which the
                             endpoint is trying to join
        :param network:      the Neutron network which the endpoint is trying
                             to join
        :returns: the tuple of stdout and stderr returned by
                  processutils.execute invoked with the executable script for
                  unbinding
        :raises: kuryr.lib.exceptions.BindingNotSupportedFailure
                 processutils.ProcessExecutionError
        """
        binding_driver = 'kuryr.lib.binding.drivers.hw_veb'
        pci_addr = neutron_port[const.BINDING_PROFILE]['pci_slot']
        pf_ifname = get_ifname_by_pci_address(pci_addr,
                                              pf_interface=True)
        vf_num = get_vf_num_by_pci_address(pci_addr)
        _, _, (stdout, stderr) = binding.port_bind(
            endpoint_id, neutron_port, subnets, pf_ifname=pf_ifname,
            vf_num=vf_num, driver=binding_driver)
        return (stdout, stderr)

    def delete_host_iface(self, endpoint_id, neutron_port):
        """Deletes a host interface after unbinding it from the host.

        The host veth interface associated to the Neutron port will be unbound
        from its vitual bridge and deleted by delegating to the selected
        kuryr-lib driver.

        :param endpoint_id:  the ID of the Docker container as string
        :param neutron_port: a port dictionary returned from
               python-neutronclient
        :returns: the tuple of stdout and stderr returned by
                  processutils.execute invoked with the executable script for
                  unbinding
        :raises: processutils.ProcessExecutionError
        """
        binding_driver = 'kuryr.lib.binding.drivers.hw_veb'
        pci_addr = neutron_port[const.BINDING_PROFILE]['pci_slot']
        pf_ifname = get_ifname_by_pci_address(pci_addr,
                                              pf_interface=True)
        vf_num = get_vf_num_by_pci_address(pci_addr)
        return binding.port_unbind(endpoint_id, neutron_port,
                                   pf_ifname=pf_ifname,
                                   vf_num=vf_num, driver=binding_driver)

    def get_container_iface_name(self, neutron_port):
        """Returns interface name of a container in the default namespace.

        :param neutron_port_id: The ID of a neutron port as string
        :returns: interface name as string
        """
        pci_addr = neutron_port[const.BINDING_PROFILE]['pci_slot']
        vf_ifname = get_ifname_by_pci_address(pci_addr)
        return vf_ifname
